"""Test that images and charts are preserved when importing sheets.

This test verifies that workbooks with images (photos/pictures) and charts
are correctly copied when using import_sheet() functionality.
"""

from __future__ import annotations

import tempfile
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Any

import openpyxl
import pytest
from openpyxl.chart import BarChart, Reference
from openpyxl.drawing.image import Image

import xpyxl as x

if TYPE_CHECKING:
    from openpyxl.worksheet.worksheet import Worksheet

# Skip tests if Pillow is not available (required for image handling)
PIL_AVAILABLE = True
PILImage: Any = None
try:
    from PIL import Image as _PILImage

    PILImage = _PILImage
except ImportError:
    PIL_AVAILABLE = False


def _get_images(ws: object) -> list[Any]:
    """Get images from worksheet (works around type annotation issues)."""
    return list(getattr(ws, "_images", []))


def _get_charts(ws: object) -> list[Any]:
    """Get charts from worksheet (works around type annotation issues)."""
    return list(getattr(ws, "_charts", []))


def _create_test_image() -> BytesIO:
    """Create a simple test PNG image."""
    if not PIL_AVAILABLE or PILImage is None:
        pytest.skip("Pillow not available")
    pil_img = PILImage.new("RGB", (100, 100), color="red")
    img_bytes = BytesIO()
    pil_img.save(img_bytes, format="PNG")
    img_bytes.seek(0)
    return img_bytes


def _create_template_with_image(path: Path) -> None:
    """Create a template workbook containing an image."""
    img_bytes = _create_test_image()

    wb = openpyxl.Workbook()
    ws: Worksheet = wb.active  # type: ignore[assignment]
    ws.title = "ImageSheet"
    ws["A1"] = "Sheet with image"

    # Add image at B2
    img = Image(img_bytes)
    img.anchor = "B2"
    ws.add_image(img)

    wb.save(path)


def _create_template_with_chart(path: Path) -> None:
    """Create a template workbook containing a chart."""
    wb = openpyxl.Workbook()
    ws: Worksheet = wb.active  # type: ignore[assignment]
    ws.title = "ChartSheet"

    # Add data for chart
    for i in range(1, 11):
        ws.cell(row=i, column=1, value=f"Item {i}")
        ws.cell(row=i, column=2, value=i * 10)

    # Create bar chart
    chart = BarChart()
    chart.title = "Test Chart"
    data = Reference(ws, min_col=2, min_row=1, max_row=10)
    categories = Reference(ws, min_col=1, min_row=1, max_row=10)
    chart.add_data(data)
    chart.set_categories(categories)
    ws.add_chart(chart, "D1")

    wb.save(path)


def _create_template_with_both(path: Path) -> None:
    """Create a template workbook containing both image and chart."""
    img_bytes = _create_test_image()

    wb = openpyxl.Workbook()
    ws: Worksheet = wb.active  # type: ignore[assignment]
    ws.title = "MixedSheet"
    ws["A1"] = "Sheet with image and chart"

    # Add image at B2
    img = Image(img_bytes)
    img.anchor = "B2"
    ws.add_image(img)

    # Add data for chart
    for i in range(1, 6):
        ws.cell(row=i + 10, column=1, value=f"Value {i}")
        ws.cell(row=i + 10, column=2, value=i * 5)

    # Create bar chart
    chart = BarChart()
    chart.title = "Sample Chart"
    data = Reference(ws, min_col=2, min_row=11, max_row=15)
    chart.add_data(data)
    ws.add_chart(chart, "D15")

    wb.save(path)


@pytest.mark.skipif(not PIL_AVAILABLE, reason="Pillow not installed")
def test_import_sheet_preserves_images() -> None:
    """Test that images are preserved when importing a sheet."""
    with tempfile.TemporaryDirectory() as tmpdir:
        tmppath = Path(tmpdir)
        template_path = tmppath / "template_with_image.xlsx"
        output_path = tmppath / "output.xlsx"

        # Create template with image
        _create_template_with_image(template_path)

        # Verify template has image
        template_wb = openpyxl.load_workbook(template_path)
        template_ws = template_wb["ImageSheet"]
        assert len(_get_images(template_ws)) == 1, "Template should have 1 image"

        # Build workbook with imported sheet
        workbook = x.workbook()[
            x.sheet("Generated")[x.row()["Hello"]],
            x.import_sheet(template_path, "ImageSheet", name="ImportedImage"),
        ]

        # Save with hybrid engine
        workbook.save(output_path, engine="hybrid")

        # Verify the output has the image
        result_wb = openpyxl.load_workbook(output_path)
        imported_ws = result_wb["ImportedImage"]

        assert len(_get_images(imported_ws)) == 1, "Imported sheet should have 1 image"
        assert imported_ws["A1"].value == "Sheet with image"  # type: ignore[union-attr]


@pytest.mark.skipif(not PIL_AVAILABLE, reason="Pillow not installed")
def test_import_sheet_preserves_charts() -> None:
    """Test that charts are preserved when importing a sheet."""
    with tempfile.TemporaryDirectory() as tmpdir:
        tmppath = Path(tmpdir)
        template_path = tmppath / "template_with_chart.xlsx"
        output_path = tmppath / "output.xlsx"

        # Create template with chart
        _create_template_with_chart(template_path)

        # Verify template has chart
        template_wb = openpyxl.load_workbook(template_path)
        template_ws = template_wb["ChartSheet"]
        assert len(_get_charts(template_ws)) == 1, "Template should have 1 chart"

        # Build workbook with imported sheet
        workbook = x.workbook()[
            x.sheet("Generated")[x.row()["Hello"]],
            x.import_sheet(template_path, "ChartSheet", name="ImportedChart"),
        ]

        # Save with hybrid engine
        workbook.save(output_path, engine="hybrid")

        # Verify the output has the chart
        result_wb = openpyxl.load_workbook(output_path)
        imported_ws = result_wb["ImportedChart"]

        assert len(_get_charts(imported_ws)) == 1, "Imported sheet should have 1 chart"
        # Verify data is preserved
        assert imported_ws["A1"].value == "Item 1"  # type: ignore[union-attr]
        assert imported_ws["B1"].value == 10  # type: ignore[union-attr]


@pytest.mark.skipif(not PIL_AVAILABLE, reason="Pillow not installed")
def test_import_sheet_preserves_images_and_charts() -> None:
    """Test that both images and charts are preserved when importing a sheet."""
    with tempfile.TemporaryDirectory() as tmpdir:
        tmppath = Path(tmpdir)
        template_path = tmppath / "template_with_both.xlsx"
        output_path = tmppath / "output.xlsx"

        # Create template with both image and chart
        _create_template_with_both(template_path)

        # Verify template has both
        template_wb = openpyxl.load_workbook(template_path)
        template_ws = template_wb["MixedSheet"]
        assert len(_get_images(template_ws)) == 1, "Template should have 1 image"
        assert len(_get_charts(template_ws)) == 1, "Template should have 1 chart"

        # Build workbook with imported sheet
        workbook = x.workbook()[
            x.import_sheet(template_path, "MixedSheet", name="ImportedMixed"),
        ]

        # Save with hybrid engine
        workbook.save(output_path, engine="hybrid")

        # Verify the output has both
        result_wb = openpyxl.load_workbook(output_path)
        imported_ws = result_wb["ImportedMixed"]

        assert len(_get_images(imported_ws)) == 1, "Imported sheet should have 1 image"
        assert len(_get_charts(imported_ws)) == 1, "Imported sheet should have 1 chart"
        assert imported_ws["A1"].value == "Sheet with image and chart"  # type: ignore[union-attr]


@pytest.mark.skipif(not PIL_AVAILABLE, reason="Pillow not installed")
def test_import_sheet_preserves_multiple_images() -> None:
    """Test that multiple images are preserved when importing a sheet."""
    with tempfile.TemporaryDirectory() as tmpdir:
        tmppath = Path(tmpdir)
        template_path = tmppath / "template_multi_image.xlsx"
        output_path = tmppath / "output.xlsx"

        # Create template with multiple images
        wb = openpyxl.Workbook()
        ws: Worksheet = wb.active  # type: ignore[assignment]
        ws.title = "MultiImage"
        ws["A1"] = "Multiple images"

        # Add 3 images at different positions
        for _i, anchor in enumerate(["B2", "D2", "F2"]):
            img_bytes = _create_test_image()
            img = Image(img_bytes)
            img.anchor = anchor
            ws.add_image(img)

        wb.save(template_path)

        # Verify template has 3 images
        template_wb = openpyxl.load_workbook(template_path)
        template_ws = template_wb["MultiImage"]
        assert len(_get_images(template_ws)) == 3, "Template should have 3 images"

        # Build workbook with imported sheet
        workbook = x.workbook()[
            x.import_sheet(template_path, "MultiImage", name="ImportedMulti"),
        ]

        # Save with hybrid engine
        workbook.save(output_path, engine="hybrid")

        # Verify the output has all 3 images
        result_wb = openpyxl.load_workbook(output_path)
        imported_ws = result_wb["ImportedMulti"]

        assert len(_get_images(imported_ws)) == 3, "Imported sheet should have 3 images"


@pytest.mark.skipif(not PIL_AVAILABLE, reason="Pillow not installed")
def test_import_sheet_with_openpyxl_engine() -> None:
    """Test that images are preserved when using openpyxl engine directly."""
    with tempfile.TemporaryDirectory() as tmpdir:
        tmppath = Path(tmpdir)
        template_path = tmppath / "template_with_image.xlsx"
        output_path = tmppath / "output.xlsx"

        # Create template with image
        _create_template_with_image(template_path)

        # Build workbook with imported sheet
        workbook = x.workbook()[
            x.import_sheet(template_path, "ImageSheet", name="ImportedImage"),
        ]

        # Save with openpyxl engine
        workbook.save(output_path, engine="openpyxl")

        # Verify the output has the image
        result_wb = openpyxl.load_workbook(output_path)
        imported_ws = result_wb["ImportedImage"]

        assert len(_get_images(imported_ws)) == 1, "Imported sheet should have 1 image"


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""Hybrid rendering engine combining xlsxwriter speed with openpyxl features.

This engine uses xlsxwriter for fast generation of new sheets and openpyxl
for importing existing sheets and performing the final save.
"""

from __future__ import annotations

from io import BytesIO
from typing import BinaryIO

from openpyxl import Workbook as _OpenpyxlWorkbook
from openpyxl import load_workbook as _load_workbook

from .base import EffectiveStyle, Engine, SaveTarget
from .openpyxl_engine import OpenpyxlEngine
from .xlsxwriter_engine import XlsxWriterEngine

__all__ = ["HybridEngine"]


class HybridEngine(Engine):
    """Hybrid engine combining xlsxwriter performance with openpyxl features.

    This engine uses xlsxwriter internally for fast generation of new sheets,
    while supporting import_sheet via openpyxl. The final workbook is saved
    using openpyxl to merge generated and imported sheets.

    Use this engine when you need both:
    - Fast generation of large workbooks (xlsxwriter)
    - Ability to import sheets from existing workbooks (openpyxl)
    """

    def __init__(self) -> None:
        super().__init__()
        self._xlsx_engine = XlsxWriterEngine()
        # Deferred import operations: (source, source_sheet_name, dest_name, show_gridlines)
        self._imports: list[
            tuple[SaveTarget | bytes | BinaryIO, str, str, bool | None]
        ] = []
        # Track sheet names in declaration order for reordering
        self._sheet_order: list[str] = []
        # Track if we have any generated sheets
        self._has_generated_sheets = False

    def create_sheet(self, name: str, show_gridlines: bool = True) -> None:
        """Create a new worksheet, delegating to xlsxwriter."""
        self._xlsx_engine.create_sheet(name, show_gridlines=show_gridlines)
        self._sheet_order.append(name)
        self._has_generated_sheets = True

    def write_cell(
        self,
        row: int,
        col: int,
        value: object,
        style: EffectiveStyle,
        border_fallback_color: str,
    ) -> None:
        """Write a cell value, delegating to xlsxwriter."""
        self._xlsx_engine.write_cell(row, col, value, style, border_fallback_color)

    def write_merged_cell(
        self,
        row: int,
        col: int,
        rowspan: int,
        colspan: int,
        value: object,
        style: EffectiveStyle,
        border_fallback_color: str,
    ) -> None:
        """Write a merged cell, delegating to xlsxwriter."""
        self._xlsx_engine.write_merged_cell(
            row,
            col,
            rowspan,
            colspan,
            value,
            style,
            border_fallback_color,
        )

    def set_column_width(self, col: int, width: float) -> None:
        """Set column width, delegating to xlsxwriter."""
        self._xlsx_engine.set_column_width(col, width)

    def set_row_height(self, row: int, height: float) -> None:
        """Set row height, delegating to xlsxwriter."""
        self._xlsx_engine.set_row_height(row, height)

    def fill_background(
        self,
        color: str,
        max_row: int,
        max_col: int,
    ) -> None:
        """Fill background color, delegating to xlsxwriter."""
        self._xlsx_engine.fill_background(color, max_row, max_col)

    def copy_sheet(
        self,
        source: SaveTarget | bytes | BinaryIO,
        sheet_name: str,
        dest_name: str,
        show_gridlines: bool | None = None,
    ) -> None:
        """Record a deferred import operation to be applied at save time.

        Unlike xlsxwriter which cannot import sheets, HybridEngine defers
        these operations and applies them using openpyxl during save().
        """
        self._imports.append((source, sheet_name, dest_name, show_gridlines))
        self._sheet_order.append(dest_name)

    def _build_openpyxl_workbook_from_xlsx(self) -> _OpenpyxlWorkbook:
        """Build the base openpyxl workbook from xlsxwriter output.

        If there are generated sheets, saves xlsxwriter to bytes and loads
        them with openpyxl. If there are no generated sheets, creates an
        empty openpyxl workbook.
        """
        if self._has_generated_sheets:
            xlsx_bytes = self._xlsx_engine.save(None)
            assert xlsx_bytes is not None
            return _load_workbook(
                BytesIO(xlsx_bytes),
                data_only=False,
                rich_text=True,
            )
        else:
            # No generated sheets, create empty workbook
            workbook = _OpenpyxlWorkbook()
            default_sheet = workbook.active
            if default_sheet is not None:
                workbook.remove(default_sheet)
            return workbook

    def _reorder_sheets(self, workbook: _OpenpyxlWorkbook) -> None:
        """Reorder workbook sheets to match the declaration order.

        This ensures sheets appear in the same order as they were declared
        in the workbook definition, regardless of whether they are generated
        or imported.
        """
        expected_order = self._sheet_order

        # Access internal _sheets list (openpyxl doesn't expose a public reorder API).
        # Reorder in-place rather than replacing the list, to avoid breaking internal
        # workbook invariants that Excel is sensitive to.
        sheets = workbook._sheets  # type: ignore[attr-defined]
        title_to_index = {ws.title: i for i, ws in enumerate(sheets)}

        insert_at = 0
        for title in expected_order:
            idx = title_to_index.get(title)
            if idx is None:
                continue

            if idx != insert_at:
                ws = sheets.pop(idx)
                sheets.insert(insert_at, ws)

                # Update indices for the moved slice.
                start = min(insert_at, idx)
                end = max(insert_at, idx)
                for j in range(start, end + 1):
                    title_to_index[sheets[j].title] = j

            insert_at += 1

        # Ensure the active sheet index is valid after reordering.
        try:
            if sheets:
                workbook.active = 0
        except Exception:
            pass

    def save(self, target: SaveTarget | None = None) -> bytes | None:
        """Finalize and save the workbook.

        This method:
        1. Phase A: Builds the base openpyxl workbook from xlsxwriter output.
        2. Phase B: Applies deferred import operations via OpenpyxlEngine.
        3. Phase C: Reorders sheets to match declaration order.
        4. Phase D: Saves the final workbook.
        """
        # Phase A: Build base workbook from xlsxwriter output
        merged_wb = self._build_openpyxl_workbook_from_xlsx()

        # Phase B: Apply imports via OpenpyxlEngine
        openpyxl_engine = OpenpyxlEngine.from_workbook(merged_wb)

        for source, sheet_name, dest_name, show_gridlines in self._imports:
            openpyxl_engine.copy_sheet(
                source,
                sheet_name,
                dest_name,
                show_gridlines=show_gridlines,
            )

        # Phase C: Reorder sheets to match declaration order
        self._reorder_sheets(merged_wb)

        # Phase D: Save the final workbook
        return openpyxl_engine.save(target)



"""XlsxWriter rendering engine implementation."""

from __future__ import annotations

from datetime import date, datetime, time, timedelta
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO

import xlsxwriter

from ..styles import normalize_hex
from .base import EffectiveStyle, Engine, SaveTarget

if TYPE_CHECKING:
    from xlsxwriter.format import Format
    from xlsxwriter.worksheet import Worksheet

__all__ = ["XlsxWriterEngine"]


# Map openpyxl border style names to xlsxwriter border style indices
# https://xlsxwriter.readthedocs.io/format.html#set_border
BORDER_STYLE_MAP: dict[str, int] = {
    "none": 0,
    "thin": 1,
    "medium": 2,
    "dashed": 3,
    "dotted": 4,
    "thick": 5,
    "double": 6,
    "hair": 7,
    "mediumDashed": 8,
    "dashDot": 9,
    "mediumDashDot": 10,
    "dashDotDot": 11,
    "mediumDashDotDot": 12,
    "slantDashDot": 13,
}


class XlsxWriterEngine(Engine):
    """Rendering engine using xlsxwriter."""

    def __init__(self) -> None:
        super().__init__()
        self._buffer = BytesIO()
        # Convert NaN/INF to Excel errors to avoid xlsxwriter write_number() failures
        workbook_options = {"in_memory": True, "nan_inf_to_errors": True}
        self._workbook = xlsxwriter.Workbook(self._buffer, workbook_options)
        self._current_sheet: Worksheet | None = None
        # Cache format objects to avoid duplicates
        self._format_cache: dict[tuple[Any, ...], Format] = {}
        self._closed = False

    def create_sheet(self, name: str, show_gridlines: bool = True) -> None:
        self._current_sheet = self._workbook.add_worksheet(name)
        self._current_sheet.hide_gridlines(0 if show_gridlines else 2)

    def _get_format(
        self, style: EffectiveStyle, border_fallback_color: str
    ) -> "Format":
        """Get or create a format object for the given style."""
        # Create a hashable key from style properties
        cache_key = (
            style.font_name,
            style.font_size,
            style.bold,
            style.italic,
            style.text_color,
            style.fill_color,
            style.horizontal_align,
            style.vertical_align,
            style.indent,
            style.wrap_text,
            style.shrink_to_fit,
            style.number_format,
            style.border,
            style.border_color or border_fallback_color if style.border else None,
            style.border_top,
            style.border_bottom,
            style.border_left,
            style.border_right,
        )

        if cache_key in self._format_cache:
            return self._format_cache[cache_key]

        fmt = self._workbook.add_format()

        # Font settings
        fmt.set_font_name(style.font_name)
        fmt.set_font_size(int(style.font_size))
        if style.bold:
            fmt.set_bold()
        if style.italic:
            fmt.set_italic()

        # Text color (convert from #RRGGBB to xlsxwriter format)
        text_color = normalize_hex(style.text_color)
        fmt.set_font_color(text_color)

        # Fill color
        if style.fill_color:
            fill_color = normalize_hex(style.fill_color)
            # XlsxWriter requires a fill pattern for background colors to render.
            # Use a solid pattern and set both foreground/background for compatibility.
            fmt.set_pattern(1)
            fmt.set_fg_color(fill_color)
            fmt.set_bg_color(fill_color)

        # Alignment
        if style.horizontal_align:
            fmt.set_align(style.horizontal_align)  # type: ignore[arg-type]
        if style.vertical_align:
            # xlsxwriter uses 'vcenter' instead of 'center' for vertical
            valign = style.vertical_align
            if valign == "center":
                valign = "vcenter"
            fmt.set_align(valign)  # type: ignore[arg-type]
        elif style.horizontal_align or style.wrap_text or style.shrink_to_fit:
            # Default vertical alignment to bottom like openpyxl
            fmt.set_align("bottom")

        if style.indent is not None:
            fmt.set_indent(style.indent)
        if style.wrap_text:
            fmt.set_text_wrap()
        if style.shrink_to_fit:
            fmt.set_shrink()

        # Number format
        if style.number_format:
            fmt.set_num_format(style.number_format)

        # Borders
        if style.border and style.border != "none":
            border_style = BORDER_STYLE_MAP.get(style.border, 1)  # default to thin
            border_color = normalize_hex(style.border_color or border_fallback_color)

            explicit = (
                style.border_top
                or style.border_bottom
                or style.border_left
                or style.border_right
            )

            if explicit:
                if style.border_top:
                    fmt.set_top(border_style)
                    fmt.set_top_color(border_color)
                if style.border_bottom:
                    fmt.set_bottom(border_style)
                    fmt.set_bottom_color(border_color)
                if style.border_left:
                    fmt.set_left(border_style)
                    fmt.set_left_color(border_color)
                if style.border_right:
                    fmt.set_right(border_style)
                    fmt.set_right_color(border_color)
            else:
                # Apply to all sides
                fmt.set_border(border_style)
                fmt.set_border_color(border_color)

        self._format_cache[cache_key] = fmt
        return fmt

    def write_cell(
        self,
        row: int,
        col: int,
        value: object,
        style: EffectiveStyle,
        border_fallback_color: str,
    ) -> None:
        if self._current_sheet is None:
            raise RuntimeError("No sheet created. Call create_sheet() first.")

        # Convert from 1-based to 0-based indexing
        row_idx = row - 1
        col_idx = col - 1

        fmt = self._get_format(style, border_fallback_color)

        # xlsxwriter has different write methods for different types
        if value is None:
            self._current_sheet.write_blank(row_idx, col_idx, None, fmt)
        elif isinstance(value, bool):
            self._current_sheet.write_boolean(row_idx, col_idx, value, fmt)
        elif isinstance(value, (date, datetime, time, timedelta)):
            self._current_sheet.write(row_idx, col_idx, value, fmt)
        elif isinstance(value, (int, float)):
            self._current_sheet.write_number(row_idx, col_idx, value, fmt)
        else:
            self._current_sheet.write_string(row_idx, col_idx, str(value), fmt)

    def write_merged_cell(
        self,
        row: int,
        col: int,
        rowspan: int,
        colspan: int,
        value: object,
        style: EffectiveStyle,
        border_fallback_color: str,
    ) -> None:
        if rowspan == 1 and colspan == 1:
            self.write_cell(row, col, value, style, border_fallback_color)
            return

        if self._current_sheet is None:
            raise RuntimeError("No sheet created. Call create_sheet() first.")

        fmt = self._get_format(style, border_fallback_color)
        self._current_sheet.merge_range(
            row - 1,
            col - 1,
            row + rowspan - 2,
            col + colspan - 2,
            value,
            fmt,
        )

    def set_column_width(self, col: int, width: float) -> None:
        if self._current_sheet is None:
            raise RuntimeError("No sheet created. Call create_sheet() first.")

        # Convert from 1-based to 0-based indexing
        col_idx = col - 1
        self._current_sheet.set_column(col_idx, col_idx, max(width, 8.0))

    def set_row_height(self, row: int, height: float) -> None:
        if self._current_sheet is None:
            raise RuntimeError("No sheet created. Call create_sheet() first.")

        # Convert from 1-based to 0-based indexing
        row_idx = row - 1
        self._current_sheet.set_row(row_idx, height)

    def fill_background(
        self,
        color: str,
        max_row: int,
        max_col: int,
    ) -> None:
        if self._current_sheet is None:
            raise RuntimeError("No sheet created. Call create_sheet() first.")

        # Create a format with just the background color
        fill_color = normalize_hex(color)
        bg_fmt = self._workbook.add_format({"bg_color": fill_color})

        # Fill all cells in the range (0-based indexing)
        for row_idx in range(max_row):
            for col_idx in range(max_col):
                self._current_sheet.write_blank(row_idx, col_idx, None, bg_fmt)

    def copy_sheet(
        self,
        source: SaveTarget | bytes | BinaryIO,
        sheet_name: str,
        dest_name: str,
        show_gridlines: bool | None = None,
    ) -> None:
        msg = (
            "import_sheet is not supported with engine='xlsxwriter'. "
            "Use engine='hybrid' or engine='openpyxl' instead."
        )
        raise NotImplementedError(msg)

    def save(self, target: SaveTarget | None = None) -> bytes | None:
        if not self._closed:
            self._workbook.close()
            self._closed = True

        data = self._buffer.getvalue()
        if target is None:
            return data

        if isinstance(target, (str, Path)):
            Path(target).write_bytes(data)
        else:
            target.write(data)
            target.flush()
        return None



from __future__ import annotations

import tempfile
from pathlib import Path

import openpyxl
import pytest

import xpyxl as x


@pytest.mark.parametrize("engine", ["openpyxl", "xlsxwriter", "hybrid"])
def test_generated_sheet_can_hide_gridlines(engine: str) -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        output_path = Path(tmpdir) / f"{engine}.xlsx"

        workbook = x.workbook()[
            x.sheet("Hidden", show_gridlines=False)[x.row()["A", "B"]]
        ]
        workbook.save(output_path, engine=engine)  # type: ignore[arg-type]

        ws = openpyxl.load_workbook(output_path)["Hidden"]
        assert ws.sheet_view.showGridLines is False


def test_cell_builder_rejects_invalid_colspan() -> None:
    with pytest.raises(ValueError, match="Cell colspan must be >= 1"):
        x.cell(colspan=0)["bad"]


def test_cell_builder_rejects_invalid_rowspan() -> None:
    with pytest.raises(ValueError, match="Cell rowspan must be >= 1"):
        x.cell(rowspan=0)["bad"]


def test_raw_scalar_row_cells_remain_unmerged() -> None:
    workbook = x.workbook()[x.sheet("Plain")[x.row()["A", "B", "C"]]]

    result = workbook.to_openpyxl()
    ws = result["Plain"]

    assert list(ws.merged_cells.ranges) == []
    assert ws["A1"].value == "A"
    assert ws["B1"].value == "B"
    assert ws["C1"].value == "C"


@pytest.mark.parametrize("engine", ["openpyxl", "xlsxwriter", "hybrid"])
def test_generated_merged_cells_round_trip(engine: str) -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        output_path = Path(tmpdir) / f"{engine}.xlsx"

        workbook = x.workbook()[
            x.sheet("Merged")[
                x.row()[
                    x.cell(style=[x.bold, x.bg_warning, x.text_center], colspan=3)[
                        "Merged Title"
                    ]
                ],
                x.row()[
                    x.cell(style=[x.bg_info], rowspan=2)["Region"],
                    "Q1",
                    "Q2",
                ],
                x.row()["North", 10, 20],
            ]
        ]

        workbook.save(output_path, engine=engine)  # type: ignore[arg-type]

        result = openpyxl.load_workbook(output_path)
        ws = result["Merged"]
        merged_ranges = {str(merged_range) for merged_range in ws.merged_cells.ranges}

        assert merged_ranges == {"A1:C1", "A2:A3"}
        assert ws["A1"].value == "Merged Title"
        assert ws["A1"].font.bold
        assert ws["A2"].value == "Region"
        assert ws["B2"].value == "Q1"
        assert ws["C2"].value == "Q2"
        assert ws["B3"].value == "North"
        assert ws["C3"].value == 10
        assert ws["D3"].value == 20


def test_rowspan_conflict_with_block_content_raises() -> None:
    workbook = x.workbook()[
        x.sheet("Conflict")[
            x.row()[x.cell(rowspan=2)["Merged"]],
            x.col()["Overlaps"],
        ]
    ]

    with pytest.raises(ValueError, match="Merged cells cannot overlap existing content"):
        workbook.save(engine="openpyxl")


def test_merged_cells_inside_tables_raise() -> None:
    workbook = x.workbook()[
        x.sheet("Table")[
            x.table()[[x.row()[x.cell(colspan=2)["Not allowed"]]]],
        ]
    ]

    with pytest.raises(ValueError, match="Merged cells are not supported inside tables"):
        workbook.save(engine="openpyxl")


def test_generated_html_renders_colspan_and_rowspan() -> None:
    workbook = x.workbook()[
        x.sheet("Merged")[
            x.row()[x.cell(colspan=3)["Merged Title"]],
            x.row()[x.cell(rowspan=2)["Region"], "Q1"],
            x.row()["North"],
        ]
    ]

    result = workbook.save(engine="html")
    assert isinstance(result, bytes)
    html = result.decode("utf-8")

    assert 'colspan="3"' in html
    assert 'rowspan="2"' in html
    assert html.count("Merged Title") == 1
    assert html.count("Region") == 1


def test_wrapping_merged_row_in_vstack_preserves_layout() -> None:
    direct = x.workbook()[
        x.sheet("Direct")[
            x.row()[x.cell(rowspan=2)["Region"], "Q1"],
            x.row()["North"],
        ]
    ].to_openpyxl()["Direct"]

    wrapped = x.workbook()[
        x.sheet("Wrapped")[
            x.vstack(x.row()[x.cell(rowspan=2)["Region"], "Q1"]),
            x.row()["North"],
        ]
    ].to_openpyxl()["Wrapped"]

    for row in range(1, 4):
        for col in range(1, 4):
            assert wrapped.cell(row=row, column=col).value == direct.cell(
                row=row, column=col
            ).value


def test_explicit_row_height_applies_to_each_row_in_rowspan() -> None:
    workbook = x.workbook()[
        x.sheet("Heights")[
            x.row()[x.cell(style=[x.row_height(40)], rowspan=2)["Merged"]],
            x.row()["Next"],
        ]
    ]

    ws = workbook.to_openpyxl()["Heights"]

    assert ws.row_dimensions[1].height == 40
    assert ws.row_dimensions[2].height == 40


def test_imported_html_renders_existing_merged_ranges() -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        source_path = Path(tmpdir) / "source.xlsx"

        wb = openpyxl.Workbook()
        ws = wb.active
        if ws is None:
            raise RuntimeError("Expected an active worksheet")
        ws.title = "Source"
        ws["A1"] = "Merged Title"
        ws.merge_cells("A1:C1")
        ws["A2"] = "Region"
        ws.merge_cells("A2:A3")
        ws["B2"] = "Q1"
        ws["B3"] = "North"
        wb.save(source_path)

        workbook = x.workbook()[x.import_sheet(source_path, "Source", name="Imported")]
        result = workbook.save(engine="html")

        assert isinstance(result, bytes)
        html = result.decode("utf-8")
        assert 'colspan="3"' in html
        assert 'rowspan="2"' in html
        assert "Merged Title" in html
        assert "Region" in html


@pytest.mark.parametrize("engine", ["openpyxl", "hybrid"])
def test_imported_sheet_can_override_gridlines(engine: str) -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        source_path = Path(tmpdir) / "source.xlsx"
        output_path = Path(tmpdir) / f"{engine}.xlsx"

        source_wb = openpyxl.Workbook()
        source_ws = source_wb.active
        if source_ws is None:
            raise RuntimeError("Expected an active worksheet")
        source_ws.title = "Source"
        source_ws.sheet_view.showGridLines = True
        source_ws["A1"] = "Imported"
        source_wb.save(source_path)

        workbook = x.workbook()[
            x.import_sheet(
                source_path,
                "Source",
                name="Imported",
                show_gridlines=False,
            )
        ]
        workbook.save(output_path, engine=engine)  # type: ignore[arg-type]

        ws = openpyxl.load_workbook(output_path)["Imported"]
        assert ws.sheet_view.showGridLines is False



"""OpenPyXL rendering engine implementation."""

from __future__ import annotations

import copy
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO

from openpyxl import Workbook, load_workbook
from openpyxl.cell.cell import MergedCell
from openpyxl.styles import Alignment, Border, Font, PatternFill, Side
from openpyxl.utils import get_column_letter

from ..styles import to_argb
from .base import EffectiveStyle, Engine, SaveTarget

if TYPE_CHECKING:
    from openpyxl.worksheet.worksheet import Worksheet

__all__ = ["OpenpyxlEngine"]


class OpenpyxlEngine(Engine):
    """Rendering engine using openpyxl."""

    def __init__(self) -> None:
        super().__init__()
        self._workbook = Workbook()
        # Remove default sheet created by openpyxl
        default_sheet = self._workbook.active
        if default_sheet is not None:
            self._workbook.remove(default_sheet)
        self._init_instance_vars()

    def _init_instance_vars(self) -> None:
        """Initialize common instance variables used by both __init__ and from_workbook."""
        self._current_sheet: Worksheet | None = None
        # Cache style objects to avoid duplicates
        self._style_cache: dict[tuple[Any, ...], dict[str, Any]] = {}
        # Cache color conversions
        self._color_cache: dict[str, str] = {}
        # Cache column letters
        self._column_letter_cache: dict[int, str] = {}

    @classmethod
    def from_workbook(cls, workbook: Workbook) -> "OpenpyxlEngine":
        """Create an engine instance wrapping an existing openpyxl Workbook.

        This allows reusing copy_sheet() and other engine methods against
        a workbook that was loaded or created externally (e.g., from xlsxwriter bytes).
        """
        engine = cls.__new__(cls)
        engine._workbook = workbook
        engine._init_instance_vars()
        return engine

    def create_sheet(self, name: str, show_gridlines: bool = True) -> None:
        self._current_sheet = self._workbook.create_sheet(title=name)
        self._current_sheet.sheet_view.showGridLines = show_gridlines

    def write_cell(
        self,
        row: int,
        col: int,
        value: object,
        style: EffectiveStyle,
        border_fallback_color: str,
    ) -> None:
        if self._current_sheet is None:
            raise RuntimeError("No sheet created. Call create_sheet() first.")

        cell = self._current_sheet.cell(row=row, column=col)
        cell.value = value  # type: ignore[assignment]
        self._apply_style(cell, style, border_fallback_color)

    def write_merged_cell(
        self,
        row: int,
        col: int,
        rowspan: int,
        colspan: int,
        value: object,
        style: EffectiveStyle,
        border_fallback_color: str,
    ) -> None:
        if rowspan == 1 and colspan == 1:
            self.write_cell(row, col, value, style, border_fallback_color)
            return

        if self._current_sheet is None:
            raise RuntimeError("No sheet created. Call create_sheet() first.")

        cell = self._current_sheet.cell(row=row, column=col)
        cell.value = value  # type: ignore[assignment]
        self._apply_style(cell, style, border_fallback_color)
        self._current_sheet.merge_cells(
            start_row=row,
            start_column=col,
            end_row=row + rowspan - 1,
            end_column=col + colspan - 1,
        )

    def _get_cached_color(self, color: str) -> str:
        """Get cached ARGB color or convert and cache it."""
        if color not in self._color_cache:
            self._color_cache[color] = to_argb(color)
        return self._color_cache[color]

    def _get_cached_styles(
        self, effective: EffectiveStyle, border_fallback_color: str
    ) -> dict[str, Any]:
        """Get or create cached style objects for the given style."""
        # Create a hashable key from style properties
        cache_key = (
            effective.font_name,
            effective.font_size,
            effective.bold,
            effective.italic,
            effective.text_color,
            effective.fill_color,
            effective.horizontal_align,
            effective.vertical_align,
            effective.indent,
            effective.wrap_text,
            effective.shrink_to_fit,
            effective.number_format,
            effective.border,
            effective.border_color or border_fallback_color
            if effective.border
            else None,
            effective.border_top,
            effective.border_bottom,
            effective.border_left,
            effective.border_right,
        )

        if cache_key in self._style_cache:
            return self._style_cache[cache_key]

        # Create style objects
        text_color_argb = self._get_cached_color(effective.text_color)
        font = Font(
            name=effective.font_name,
            size=effective.font_size,
            bold=effective.bold,
            italic=effective.italic,
            color=text_color_argb,
        )

        fill: PatternFill | None = None
        if effective.fill_color:
            fill_color_argb = self._get_cached_color(effective.fill_color)
            fill = PatternFill(
                fill_type="solid",
                start_color=fill_color_argb,
                end_color=fill_color_argb,
            )

        alignment: Alignment | None = None
        align_kwargs: dict[str, Any] = {}
        if effective.horizontal_align:
            align_kwargs["horizontal"] = effective.horizontal_align
        if effective.vertical_align:
            align_kwargs["vertical"] = effective.vertical_align
        if effective.indent is not None:
            align_kwargs["indent"] = effective.indent
        if effective.wrap_text:
            align_kwargs["wrap_text"] = True
        if effective.shrink_to_fit:
            align_kwargs["shrink_to_fit"] = True
        if align_kwargs:
            align_kwargs.setdefault("vertical", "bottom")
            alignment = Alignment(**align_kwargs)  # type: ignore[arg-type]
        elif effective.wrap_text or effective.shrink_to_fit:
            alignment = Alignment(
                wrap_text=True if effective.wrap_text else None,
                shrink_to_fit=True if effective.shrink_to_fit else None,
            )

        border: Border | None = None
        if effective.border:
            border_color = effective.border_color or border_fallback_color
            border_color_argb = self._get_cached_color(border_color)

            def build_side(enabled: bool) -> Side | None:
                if not enabled:
                    return None
                return Side(style=effective.border, color=border_color_argb)

            explicit = (
                effective.border_top
                or effective.border_bottom
                or effective.border_left
                or effective.border_right
            )
            if explicit:
                border = Border(
                    left=build_side(effective.border_left),
                    right=build_side(effective.border_right),
                    top=build_side(effective.border_top),
                    bottom=build_side(effective.border_bottom),
                )
            else:
                side = build_side(True)
                border = Border(left=side, right=side, top=side, bottom=side)

        styles = {
            "font": font,
            "fill": fill,
            "alignment": alignment,
            "border": border,
            "number_format": effective.number_format,
        }

        self._style_cache[cache_key] = styles
        return styles

    def _apply_style(
        self, cell: object, effective: EffectiveStyle, border_fallback_color: str
    ) -> None:
        """Apply style to an openpyxl cell."""
        styles = self._get_cached_styles(effective, border_fallback_color)

        cell.font = styles["font"]  # type: ignore[attr-defined]

        if styles["fill"]:
            cell.fill = styles["fill"]  # type: ignore[attr-defined]

        if styles["alignment"]:
            cell.alignment = styles["alignment"]  # type: ignore[attr-defined]
        elif cell.alignment is None and (  # type: ignore[attr-defined]
            effective.wrap_text or effective.shrink_to_fit
        ):
            # Handle edge case where alignment wasn't cached but wrap/shrink is needed
            cell.alignment = Alignment(  # type: ignore[attr-defined]
                wrap_text=True if effective.wrap_text else None,
                shrink_to_fit=True if effective.shrink_to_fit else None,
            )

        if styles["number_format"]:
            cell.number_format = styles["number_format"]  # type: ignore[attr-defined]

        if styles["border"]:
            cell.border = styles["border"]  # type: ignore[attr-defined]

    def set_column_width(self, col: int, width: float) -> None:
        if self._current_sheet is None:
            raise RuntimeError("No sheet created. Call create_sheet() first.")

        if col not in self._column_letter_cache:
            self._column_letter_cache[col] = get_column_letter(col)
        letter = self._column_letter_cache[col]
        self._current_sheet.column_dimensions[letter].width = max(width, 8.0)

    def set_row_height(self, row: int, height: float) -> None:
        if self._current_sheet is None:
            raise RuntimeError("No sheet created. Call create_sheet() first.")

        self._current_sheet.row_dimensions[row].height = height

    def fill_background(
        self,
        color: str,
        max_row: int,
        max_col: int,
    ) -> None:
        if self._current_sheet is None:
            raise RuntimeError("No sheet created. Call create_sheet() first.")

        fill_color = self._get_cached_color(color)
        sheet_fill = PatternFill(
            fill_type="solid", start_color=fill_color, end_color=fill_color
        )
        # Reuse the same PatternFill object for all cells (openpyxl supports this)
        for row in self._current_sheet.iter_rows(
            min_row=1, max_row=max_row, min_col=1, max_col=max_col
        ):
            for cell in row:
                cell.fill = sheet_fill

    def _ensure_named_styles(self, source_wb: Workbook) -> None:
        """Ensure named styles used by imported sheets exist in the destination workbook."""
        try:
            source_styles = list(getattr(source_wb, "named_styles", []))
            dest_styles = list(getattr(self._workbook, "named_styles", []))
        except Exception:
            return

        existing = {getattr(style, "name", None) for style in dest_styles}
        for style in source_styles:
            name = getattr(style, "name", None)
            if not name or name in existing:
                continue
            try:
                self._workbook.add_named_style(copy.copy(style))
                existing.add(name)
            except Exception:
                # NamedStyle copying is best-effort; cell-level style objects
                # are still copied below.
                pass

    def _clone_sheet_contents(self, source_ws: Worksheet, target_ws: Worksheet) -> None:
        """Copy values and basic layout from a source sheet into a target sheet.

        This avoids copying workbook-scoped style indices (which can corrupt files
        cross-workbook) and instead copies resolved style objects.
        """
        # Basic sheet properties (safe, value-based)
        try:
            target_ws.sheet_properties = copy.copy(source_ws.sheet_properties)
        except Exception:
            pass
        try:
            target_ws.sheet_format = copy.copy(source_ws.sheet_format)
        except Exception:
            pass
        try:
            target_ws.page_margins = copy.copy(source_ws.page_margins)
        except Exception:
            pass
        try:
            target_ws.page_setup = copy.copy(source_ws.page_setup)
        except Exception:
            pass
        try:
            target_ws.print_options = copy.copy(source_ws.print_options)
        except Exception:
            pass
        try:
            target_ws.protection = copy.copy(source_ws.protection)
        except Exception:
            pass

        try:
            target_ws.freeze_panes = source_ws.freeze_panes
        except Exception:
            pass
        try:
            target_ws.auto_filter.ref = source_ws.auto_filter.ref
        except Exception:
            pass
        try:
            target_ws.sheet_view.zoomScale = source_ws.sheet_view.zoomScale
        except Exception:
            pass

        # Values (including formulas) + cell-level style objects.
        for row in source_ws.iter_rows():
            for cell in row:
                if isinstance(cell, MergedCell):
                    continue
                if cell.row is None or cell.column is None:
                    continue
                target_cell = target_ws.cell(
                    row=cell.row, column=cell.column, value=cell.value
                )
                if getattr(cell, "has_style", False):
                    # Copy resolved style objects, not style indices.
                    target_cell.font = copy.copy(cell.font)  # pyright: ignore[reportAttributeAccessIssue]
                    target_cell.fill = copy.copy(cell.fill)  # pyright: ignore[reportAttributeAccessIssue]
                    target_cell.border = copy.copy(cell.border)  # pyright: ignore[reportAttributeAccessIssue]
                    target_cell.alignment = copy.copy(cell.alignment)  # pyright: ignore[reportAttributeAccessIssue]
                    target_cell.number_format = cell.number_format
                    target_cell.protection = copy.copy(cell.protection)  # pyright: ignore[reportAttributeAccessIssue]
                if getattr(cell, "hyperlink", None):
                    target_cell._hyperlink = copy.copy(cell._hyperlink)  # type: ignore[attr-defined]
                if cell.comment:
                    target_cell.comment = copy.copy(cell.comment)

        # Merge ranges
        for merged_range in source_ws.merged_cells.ranges:
            target_ws.merge_cells(str(merged_range))

        # Data validations
        try:
            if source_ws.data_validations is not None:
                for dv in source_ws.data_validations.dataValidation:
                    cloned = copy.copy(dv)
                    cloned.ranges = list(dv.ranges)
                    target_ws.add_data_validation(cloned)
        except Exception:
            pass

        # Conditional formatting (best-effort)
        try:
            if source_ws.conditional_formatting:
                for key, rules in source_ws.conditional_formatting._cf_rules.items():  # type: ignore[attr-defined]
                    target_ws.conditional_formatting._cf_rules[key] = copy.deepcopy(  # type: ignore[attr-defined]
                        rules
                    )
        except Exception:
            pass

        def _copy_dimension_attrs(source_dim: object, dest_dim: object) -> None:
            # Copy common, non-style attributes; skip workbook-scoped style internals.
            for attr in (
                "width",
                "height",
                "hidden",
                "outlineLevel",
                "collapsed",
                "bestFit",
                "customWidth",
                "customHeight",
            ):
                try:
                    if hasattr(source_dim, attr):
                        value = getattr(source_dim, attr)
                        if value is not None:
                            setattr(dest_dim, attr, value)
                except Exception:
                    pass

        for column, dimension in source_ws.column_dimensions.items():
            _copy_dimension_attrs(dimension, target_ws.column_dimensions[column])

        for row, dimension in source_ws.row_dimensions.items():
            _copy_dimension_attrs(dimension, target_ws.row_dimensions[row])

        # Images (photos/pictures)
        self._copy_images(source_ws, target_ws)

        # Charts
        self._copy_charts(source_ws, target_ws)

    def _copy_images(self, source_ws: Worksheet, target_ws: Worksheet) -> None:
        """Copy all images from source worksheet to target worksheet."""
        try:
            from openpyxl.drawing.image import Image
        except ImportError:
            # Image support not available
            return

        for img in getattr(source_ws, "_images", []):
            try:
                # Get the image data from the ref (BytesIO containing the image data)
                ref = getattr(img, "ref", None)
                if ref is None:
                    continue

                # Seek to start if possible
                if hasattr(ref, "seek"):
                    try:
                        ref.seek(0)
                    except Exception:
                        pass

                # Create a new Image instance from the image data
                new_img = Image(ref)

                # Copy anchor (position in the sheet)
                anchor = getattr(img, "anchor", None)
                if anchor is not None:
                    new_img.anchor = anchor

                # Copy dimensions if set
                if getattr(img, "width", None) is not None:
                    new_img.width = img.width
                if getattr(img, "height", None) is not None:
                    new_img.height = img.height

                target_ws.add_image(new_img)
            except Exception:
                # Image copying is best-effort; skip failures
                pass

    def _copy_charts(self, source_ws: Worksheet, target_ws: Worksheet) -> None:
        """Copy all charts from source worksheet to target worksheet.

        When copying charts, we need to update the data references to point
        to the target sheet instead of the source sheet, since sheet names
        may differ between source and destination.
        """
        source_name = source_ws.title
        target_name = target_ws.title

        for chart in getattr(source_ws, "_charts", []):
            try:
                # Deep copy the chart to avoid sharing references
                new_chart = copy.deepcopy(chart)

                # Update sheet references in chart data
                self._update_chart_references(new_chart, source_name, target_name)

                target_ws.add_chart(new_chart)
            except Exception:
                # Chart copying is best-effort; skip failures
                pass

    def _update_chart_references(
        self, chart: object, old_sheet: str, new_sheet: str
    ) -> None:
        """Update all sheet references in a chart from old_sheet to new_sheet."""
        import re

        def update_ref(formula: str | None) -> str | None:
            if not formula:
                return formula

            # Escape special regex chars in old_sheet
            old_escaped = re.escape(old_sheet)

            # Pattern for quoted sheet name: 'Sheet Name'!
            quoted_pattern = f"'{old_escaped}'!"
            # Pattern for unquoted sheet name: SheetName!
            unquoted_pattern = f"{old_escaped}!"

            # New sheet name (quote if has spaces or special chars)
            if " " in new_sheet or any(c in new_sheet for c in "'![]:"):
                new_ref = f"'{new_sheet}'!"
            else:
                new_ref = f"{new_sheet}!"

            # Replace quoted version first
            result = re.sub(quoted_pattern, new_ref, formula)
            # Then try unquoted version if no change
            if result == formula:
                result = re.sub(unquoted_pattern, new_ref, formula)

            return result

        # Update series data references
        for series in getattr(chart, "series", []):
            # Value reference (numRef)
            if hasattr(series, "val") and series.val:
                num_ref = getattr(series.val, "numRef", None)
                if num_ref and hasattr(num_ref, "f"):
                    num_ref.f = update_ref(num_ref.f)

            # Category reference (can be numRef or strRef)
            if hasattr(series, "cat") and series.cat:
                cat_num_ref = getattr(series.cat, "numRef", None)
                if cat_num_ref and hasattr(cat_num_ref, "f"):
                    cat_num_ref.f = update_ref(cat_num_ref.f)

                cat_str_ref = getattr(series.cat, "strRef", None)
                if cat_str_ref and hasattr(cat_str_ref, "f"):
                    cat_str_ref.f = update_ref(cat_str_ref.f)

            # X values (for scatter/bubble charts)
            if hasattr(series, "xVal") and series.xVal:
                x_num_ref = getattr(series.xVal, "numRef", None)
                if x_num_ref and hasattr(x_num_ref, "f"):
                    x_num_ref.f = update_ref(x_num_ref.f)

            # Y values (for scatter/bubble charts)
            if hasattr(series, "yVal") and series.yVal:
                y_num_ref = getattr(series.yVal, "numRef", None)
                if y_num_ref and hasattr(y_num_ref, "f"):
                    y_num_ref.f = update_ref(y_num_ref.f)

    def _load_source_workbook(self, source: SaveTarget | bytes | BinaryIO) -> Workbook:
        if isinstance(source, (str, Path)):
            return load_workbook(
                filename=source,
                data_only=False,
            )

        buffer: BinaryIO
        if isinstance(source, bytes):
            buffer = BytesIO(source)
        else:
            buffer = source
            if hasattr(buffer, "seek"):
                try:
                    buffer.seek(0)
                except Exception:
                    pass

        return load_workbook(
            buffer,
            data_only=False,
        )

    def _ensure_sheet_name_available(self, name: str) -> None:
        if name in self._workbook.sheetnames:
            raise ValueError(f"Sheet '{name}' already exists in destination workbook")

    def copy_sheet(
        self,
        source: SaveTarget | bytes | BinaryIO,
        sheet_name: str,
        dest_name: str,
        show_gridlines: bool | None = None,
    ) -> None:
        """Copy a sheet from an external workbook into the current workbook.

        Always copies via _clone_sheet_contents to ensure consistent behavior
        regardless of whether imported sheets come before or after generated sheets.
        """
        source_wb = self._load_source_workbook(source)

        if sheet_name not in source_wb.sheetnames:
            raise ValueError(f"Sheet '{sheet_name}' not found in source workbook")

        self._ensure_sheet_name_available(dest_name)
        self._ensure_named_styles(source_wb)

        source_ws = source_wb[sheet_name]
        target_ws = self._workbook.create_sheet(title=dest_name)
        self._clone_sheet_contents(source_ws, target_ws)
        target_ws.sheet_view.showGridLines = (
            source_ws.sheet_view.showGridLines
            if show_gridlines is None
            else show_gridlines
        )

        self._current_sheet = target_ws

    def save(self, target: SaveTarget | None = None) -> bytes | None:
        if target is None:
            buffer = BytesIO()
            self._workbook.save(buffer)
            return buffer.getvalue()

        if isinstance(target, (str, Path)):
            self._workbook.save(str(target))
        else:
            self._workbook.save(target)
        return None



"""Pluggable rendering engines for xpyxl."""

from __future__ import annotations

from typing import Literal

from .base import Engine
from .html_engine import HtmlEngine
from .hybrid_engine import HybridEngine
from .openpyxl_engine import OpenpyxlEngine
from .xlsxwriter_engine import XlsxWriterEngine

__all__ = [
    "Engine",
    "HtmlEngine",
    "HybridEngine",
    "OpenpyxlEngine",
    "XlsxWriterEngine",
    "EngineName",
    "get_engine",
]

EngineName = Literal["openpyxl", "xlsxwriter", "hybrid", "html"]


def get_engine(name: EngineName) -> Engine:
    """Create an engine instance for the given name.

    Available engines:
    - "hybrid" (default): Combines xlsxwriter speed for generated sheets with
      openpyxl for importing sheets. Best balance of speed and features.
    - "openpyxl": Full-featured engine using openpyxl. Supports all features
      including import_sheet.
    - "xlsxwriter": Fast generation engine using xlsxwriter. Does NOT support
      import_sheet - use "hybrid" or "openpyxl" for that.
    - "html": Experimental HTML output engine.
    """
    if name == "openpyxl":
        return OpenpyxlEngine()
    elif name == "xlsxwriter":
        return XlsxWriterEngine()
    elif name == "hybrid":
        return HybridEngine()
    elif name == "html":
        return HtmlEngine()
    else:
        raise ValueError(f"Unknown engine: {name}")



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, Literal, TypedDict, Unpack

__all__ = [
    "Style",
    "combine_styles",
    "normalize_hex",
    "to_argb",
    "text_xs",
    "text_sm",
    "text_base",
    "text_lg",
    "text_xl",
    "text_2xl",
    "text_3xl",
    "bold",
    "italic",
    "mono",
    "muted",
    "text_muted",
    "text_primary",
    "text_white",
    "text_red",
    "text_green",
    "text_blue",
    "text_orange",
    "text_purple",
    "text_black",
    "text_gray",
    "text_left",
    "text_center",
    "text_right",
    "align_top",
    "align_middle",
    "align_bottom",
    "wrap",
    "nowrap",
    "wrap_shrink",
    "allow_overflow",
    "row_height",
    "row_width",
    "bg_red",
    "bg_primary",
    "bg_muted",
    "bg_success",
    "bg_warning",
    "bg_info",
    "border_all",
    "border_top",
    "border_bottom",
    "border_left",
    "border_right",
    "border_x",
    "border_y",
    "border_red",
    "border_green",
    "border_blue",
    "border_orange",
    "border_purple",
    "border_black",
    "border_gray",
    "border_white",
    "border_muted",
    "border_primary",
    "border_thin",
    "border_medium",
    "border_thick",
    "border_dashed",
    "border_dotted",
    "border_double",
    "border_none",
    "table_bordered",
    "table_banded",
    "table_compact",
    "number_comma",
    "number_precision",
    "percent",
    "currency_usd",
    "currency_eur",
    "date_short",
    "datetime_short",
    "time_short",
    "BorderStyleName",
    "BorderStyleLiteral",
]


BorderStyleLiteral = Literal[
    "dashDot",
    "dashDotDot",
    "dashed",
    "dotted",
    "double",
    "hair",
    "medium",
    "mediumDashDot",
    "mediumDashDotDot",
    "mediumDashed",
    "slantDashDot",
    "thick",
    "thin",
]
BorderStyleName = BorderStyleLiteral | Literal["none"]

DEFAULT_BORDER_STYLE_NAME: BorderStyleLiteral = "thin"


def normalize_hex(value: str) -> str:
    text = value.strip()
    if not text:
        raise ValueError("Color values cannot be empty")
    if text.startswith("#"):
        text = text[1:]
    if len(text) == 3:
        text = "".join(ch * 2 for ch in text)
    if len(text) != 6:
        raise ValueError(f"Expected 6 hex characters, got '{value}'")
    return "#" + text.upper()


def to_argb(value: str) -> str:
    rgb = normalize_hex(value)[1:]
    return "FF" + rgb


@dataclass(frozen=True)
class Style:
    name: str = ""
    font_name: str | None = None
    font_size: float | None = None
    font_size_delta: float | None = None
    bold: bool | None = None
    italic: bool | None = None
    mono: bool | None = None
    text_color: str | None = None
    fill_color: str | None = None
    horizontal_align: str | None = None
    vertical_align: str | None = None
    indent: int | None = None
    wrap_text: bool | None = None
    shrink_to_fit: bool | None = None
    auto_width: bool | None = None
    row_height: float | None = None
    row_width: float | None = None
    number_format: str | None = None
    border: BorderStyleName | None = None
    border_color: str | None = None
    border_top: bool | None = None
    border_bottom: bool | None = None
    border_left: bool | None = None
    border_right: bool | None = None
    table_banded: bool | None = None
    table_bordered: bool | None = None
    table_compact: bool | None = None

    def merge(self, other: Style) -> Style:
        base_delta = 0.0 if self.font_size_delta is None else self.font_size_delta
        other_delta = 0.0 if other.font_size_delta is None else other.font_size_delta
        merged_delta = base_delta + other_delta
        delta_value = (
            merged_delta
            if (self.font_size_delta is not None or other.font_size_delta is not None)
            else None
        )
        return Style(
            name=other.name or self.name,
            font_name=other.font_name or self.font_name,
            font_size=other.font_size
            if other.font_size is not None
            else self.font_size,
            font_size_delta=delta_value,
            bold=other.bold if other.bold is not None else self.bold,
            italic=other.italic if other.italic is not None else self.italic,
            mono=other.mono if other.mono is not None else self.mono,
            text_color=other.text_color or self.text_color,
            fill_color=other.fill_color or self.fill_color,
            horizontal_align=other.horizontal_align or self.horizontal_align,
            vertical_align=other.vertical_align or self.vertical_align,
            indent=other.indent if other.indent is not None else self.indent,
            wrap_text=other.wrap_text
            if other.wrap_text is not None
            else self.wrap_text,
            shrink_to_fit=other.shrink_to_fit
            if other.shrink_to_fit is not None
            else self.shrink_to_fit,
            auto_width=other.auto_width
            if other.auto_width is not None
            else self.auto_width,
            row_height=other.row_height
            if other.row_height is not None
            else self.row_height,
            row_width=other.row_width
            if other.row_width is not None
            else self.row_width,
            number_format=other.number_format or self.number_format,
            border=other.border if other.border is not None else self.border,
            border_color=other.border_color
            if other.border_color is not None
            else self.border_color,
            border_top=other.border_top
            if other.border_top is not None
            else self.border_top,
            border_bottom=other.border_bottom
            if other.border_bottom is not None
            else self.border_bottom,
            border_left=other.border_left
            if other.border_left is not None
            else self.border_left,
            border_right=other.border_right
            if other.border_right is not None
            else self.border_right,
            table_banded=other.table_banded
            if other.table_banded is not None
            else self.table_banded,
            table_bordered=other.table_bordered
            if other.table_bordered is not None
            else self.table_bordered,
            table_compact=other.table_compact
            if other.table_compact is not None
            else self.table_compact,
        )


def combine_styles(styles: Iterable[Style], *, base: Style | None = None) -> Style:
    combined = base or Style()
    for style in styles:
        combined = combined.merge(style)
    return combined


def _style(name: str, **kwargs: Unpack[_StyleKwargs]) -> Style:
    return Style(name=name, **kwargs)


# fmt: off
text_xs = _style("text_xs", font_size_delta=-2.0)
text_sm = _style("text_sm", font_size_delta=-1.0)
text_base = _style("text_base", font_size_delta=0.0)
text_lg = _style("text_lg", font_size_delta=1.0)
text_xl = _style("text_xl", font_size_delta=4.0)
text_2xl = _style("text_2xl", font_size_delta=6.0)
text_3xl = _style("text_3xl", font_size_delta=8.0)

bold = _style("bold", bold=True)
italic = _style("italic", italic=True)
mono = _style("mono", mono=True)

muted = _style("muted", text_color=normalize_hex("#6B7280"))
text_muted = muted
text_primary = _style("text_primary", text_color=normalize_hex("#2563EB"))
text_white = _style("text_white", text_color=normalize_hex("#FFFFFF"))
text_red = _style("text_red", text_color=normalize_hex("#DC2626"))
text_green = _style("text_green", text_color=normalize_hex("#16A34A"))
text_blue = _style("text_blue", text_color=normalize_hex("#2563EB"))
text_orange = _style("text_orange", text_color=normalize_hex("#EA580C"))
text_purple = _style("text_purple", text_color=normalize_hex("#7C3AED"))
text_black = _style("text_black", text_color=normalize_hex("#111827"))
text_gray = _style("text_gray", text_color=normalize_hex("#4B5563"))

bg_red = _style("bg_red", fill_color=normalize_hex("#F04438"))
bg_primary = _style("bg_primary", fill_color=normalize_hex("#2563EB"))
bg_muted = _style("bg_muted", fill_color=normalize_hex("#6B7280"))
bg_success = _style("bg_success", fill_color=normalize_hex("#047857"))
bg_warning = _style("bg_warning", fill_color=normalize_hex("#B45309"))
bg_info = _style("bg_info", fill_color=normalize_hex("#0EA5E9"))

text_left = _style("text_left", horizontal_align="left")
text_center = _style("text_center", horizontal_align="center")
text_right = _style("text_right", horizontal_align="right")
align_top = _style("align_top", vertical_align="top")
align_middle = _style("align_middle", vertical_align="center")
align_bottom = _style("align_bottom", vertical_align="bottom")
wrap = _style("wrap", wrap_text=True)
nowrap = _style("nowrap", wrap_text=False)
wrap_shrink = _style("wrap_shrink", wrap_text=True, shrink_to_fit=True)
allow_overflow = _style("allow_overflow", auto_width=False)


def row_height(value: float) -> Style:
    if value <= 0:
        raise ValueError("Row height must be positive")
    return Style(name=f"row_height_{value:g}", row_height=value)


def row_width(value: float) -> Style:
    if value <= 0:
        raise ValueError("Row width must be positive")
    return Style(name=f"row_width_{value:g}", row_width=value)

border_all = _style(
    "border_all",
    border=DEFAULT_BORDER_STYLE_NAME,
    border_top=True,
    border_bottom=True,
    border_left=True,
    border_right=True,
)
border_top = _style("border_top", border=DEFAULT_BORDER_STYLE_NAME, border_top=True)
border_bottom = _style(
    "border_bottom", border=DEFAULT_BORDER_STYLE_NAME, border_bottom=True
)
border_left = _style("border_left", border=DEFAULT_BORDER_STYLE_NAME, border_left=True)
border_right = _style(
    "border_right", border=DEFAULT_BORDER_STYLE_NAME, border_right=True
)
border_x = _style(
    "border_x",
    border=DEFAULT_BORDER_STYLE_NAME,
    border_left=True,
    border_right=True,
)
border_y = _style(
    "border_y",
    border=DEFAULT_BORDER_STYLE_NAME,
    border_top=True,
    border_bottom=True,
)

border_thin = _style("border_thin", border="thin")
border_medium = _style("border_medium", border="medium")
border_thick = _style("border_thick", border="thick")
border_dashed = _style("border_dashed", border="dashed")
border_dotted = _style("border_dotted", border="dotted")
border_double = _style("border_double", border="double")
border_none = _style("border_none", border="none")

border_primary = _style("border_primary", border_color=normalize_hex("#2563EB"))
border_muted = _style("border_muted", border_color=normalize_hex("#6B7280"))
border_red = _style("border_red", border_color=normalize_hex("#DC2626"))
border_green = _style("border_green", border_color=normalize_hex("#16A34A"))
border_blue = _style("border_blue", border_color=normalize_hex("#2563EB"))
border_orange = _style("border_orange", border_color=normalize_hex("#EA580C"))
border_purple = _style("border_purple", border_color=normalize_hex("#7C3AED"))
border_black = _style("border_black", border_color=normalize_hex("#111827"))
border_gray = _style("border_gray", border_color=normalize_hex("#4B5563"))
border_white = _style("border_white", border_color=normalize_hex("#FFFFFF"))

table_bordered = _style("table_bordered", table_bordered=True)
table_banded = _style("table_banded", table_banded=True)
table_compact = _style("table_compact", table_compact=True)

number_comma = _style("number_comma", number_format="#,##0")
number_precision = _style("number_precision", number_format="#,##0.00")
percent = _style("percent", number_format="0.00%")
currency_usd = _style("currency_usd", number_format="$#,##0.00")
currency_eur = _style("currency_eur", number_format="€#,##0.00")
date_short = _style("date_short", number_format="yyyy-mm-dd")
datetime_short = _style("datetime_short", number_format="yyyy-mm-dd hh:mm")
time_short = _style("time_short", number_format="hh:mm")
class _StyleKwargs(TypedDict, total=False):
    font_name: str | None
    font_size: float | None
    font_size_delta: float | None
    bold: bool | None
    italic: bool | None
    mono: bool | None
    text_color: str | None
    fill_color: str | None
    horizontal_align: str | None
    vertical_align: str | None
    indent: int | None
    wrap_text: bool | None
    shrink_to_fit: bool | None
    auto_width: bool | None
    row_height: float | None
    row_width: float | None
    number_format: str | None
    border: BorderStyleName | None
    border_color: str | None
    border_top: bool | None
    border_bottom: bool | None
    border_left: bool | None
    border_right: bool | None
    table_banded: bool | None
    table_bordered: bool | None
    table_compact: bool | None



from __future__ import annotations

from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import BinaryIO, TypeAlias, cast

from ._workbook import Workbook
from .nodes import (
    CellNode,
    CellValue,
    ColumnNode,
    HorizontalStackNode,
    ImportedSheetNode,
    RowNode,
    SheetComponent,
    SheetItem,
    SheetNode,
    SpacerNode,
    TableNode,
    VerticalStackNode,
    WorkbookNode,
)
from .styles import Style, normalize_hex

__all__ = [
    "cell",
    "col",
    "import_sheet",
    "row",
    "space",
    "vstack",
    "hstack",
    "sheet",
    "table",
    "workbook",
]


ColumnKey: TypeAlias = str
CellSource: TypeAlias = CellValue | CellNode
Node = (
    CellNode
    | RowNode
    | ColumnNode
    | TableNode
    | SpacerNode
    | VerticalStackNode
    | HorizontalStackNode
)


def _as_tuple(values: object) -> tuple[object, ...]:
    if isinstance(values, tuple):
        return values
    if isinstance(values, list):
        return tuple(values)
    if isinstance(values, Sequence) and not isinstance(values, (str, bytes, bytearray)):
        return tuple(values)
    return (values,)


def _ensure_cell(value: CellSource) -> CellNode:
    if isinstance(value, CellNode):
        return value
    if isinstance(value, (RowNode, ColumnNode, TableNode)):
        msg = "Cannot nest row/column/table directly inside a cell"
        raise TypeError(msg)
    return CellNode(value=value)


def _ensure_component(value: object) -> SheetComponent:
    if isinstance(value, Node):
        return value
    msg = "Layouts accept composed nodes. Call the primitive builder before nesting."
    raise TypeError(msg)


def _coerce_row(value: object, extra_styles: Sequence[Style] = ()) -> RowNode:
    if isinstance(value, RowNode):
        if not extra_styles:
            return value
        return RowNode(cells=value.cells, styles=tuple(extra_styles) + value.styles)
    cells = tuple(_ensure_cell(item) for item in _as_tuple(value))
    return RowNode(cells=cells, styles=tuple(extra_styles))


def _rows_from_records(
    records: Sequence[Mapping[ColumnKey, CellSource]],
    *,
    header_styles: Sequence[Style],
    column_order: Sequence[ColumnKey] | None,
) -> tuple[tuple[RowNode, ...], RowNode | None]:
    if not records and not column_order:
        return (), None

    columns: list[ColumnKey] = list(column_order or ())
    seen = set(columns)
    for record in records:
        for key in record.keys():
            if key not in seen:
                seen.add(key)
                columns.append(key)

    header_node = _coerce_row(columns, extra_styles=header_styles) if columns else None
    body_rows = tuple(
        RowNode(cells=tuple(_ensure_cell(record.get(col)) for col in columns))
        for record in records
    )
    return body_rows, header_node


def _rows_from_dict_of_lists(
    table_data: Mapping[ColumnKey, Sequence[CellSource]],
    *,
    header_styles: Sequence[Style],
    column_order: Sequence[ColumnKey] | None,
) -> tuple[tuple[RowNode, ...], RowNode | None]:
    columns: list[ColumnKey] = list(column_order or ())
    seen = set(columns)
    for name in table_data.keys():
        if name not in seen:
            seen.add(name)
            columns.append(name)

    if not columns:
        return (), None

    normalized: dict[ColumnKey, Sequence[CellSource]] = {}
    lengths: set[int] = set()
    for key, values in table_data.items():
        if isinstance(values, Mapping) or isinstance(values, (str, bytes, bytearray)):
            msg = (
                "Dict-of-lists table data must map column names to sequences of values"
            )
            raise TypeError(msg)
        if not isinstance(values, Sequence):
            values = tuple(values)
        normalized[key] = values
        lengths.add(len(values))

    if len(lengths) > 1:
        msg = f"All columns must be the same length; got lengths {sorted(lengths)}"
        raise ValueError(msg)

    row_count = lengths.pop() if lengths else 0
    body_rows: list[RowNode] = []
    for idx in range(row_count):
        body_rows.append(
            RowNode(
                cells=tuple(
                    _ensure_cell(
                        normalized[col][idx]
                        if col in normalized and idx < len(normalized[col])
                        else None
                    )
                    for col in columns
                )
            )
        )

    header_node = _coerce_row(columns, extra_styles=header_styles)
    return tuple(body_rows), header_node


class _BuilderBase:
    def __init__(self, *, styles: Sequence[Style] | None = None) -> None:
        self._styles: tuple[Style, ...] = tuple(styles or ())


class CellBuilder(_BuilderBase):
    def __init__(
        self,
        *,
        styles: Sequence[Style] | None = None,
        colspan: int = 1,
        rowspan: int = 1,
    ) -> None:
        super().__init__(styles=styles)
        if colspan < 1:
            msg = "Cell colspan must be >= 1"
            raise ValueError(msg)
        if rowspan < 1:
            msg = "Cell rowspan must be >= 1"
            raise ValueError(msg)
        self._colspan = colspan
        self._rowspan = rowspan

    def __getitem__(self, value: CellSource) -> CellNode:
        return CellNode(
            value=value,
            styles=self._styles,
            colspan=self._colspan,
            rowspan=self._rowspan,
        )


class RowBuilder(_BuilderBase):
    def __getitem__(self, values: Sequence[CellSource] | CellSource) -> RowNode:
        cells = tuple(_ensure_cell(item) for item in _as_tuple(values))
        return RowNode(cells=cells, styles=self._styles)


class ColumnBuilder(_BuilderBase):
    def __getitem__(self, values: Sequence[CellSource] | CellSource) -> ColumnNode:
        cells = tuple(_ensure_cell(item) for item in _as_tuple(values))
        return ColumnNode(cells=cells, styles=self._styles)


class TableBuilder(_BuilderBase):
    def __init__(
        self,
        *,
        styles: Sequence[Style] | None = None,
        header_style: Sequence[Style] | None = None,
        columns: Sequence[ColumnKey] | None = None,
    ) -> None:
        super().__init__(styles=styles)
        self._header_styles: tuple[Style, ...] = tuple(header_style or ())
        self._columns: tuple[ColumnKey, ...] | None = (
            tuple(columns) if columns is not None else None
        )

    def __getitem__(
        self,
        rows: Sequence[RowNode]
        | Sequence[Sequence[CellSource]]
        | Sequence[Mapping[ColumnKey, CellSource]]
        | Mapping[ColumnKey, Sequence[CellSource]],
    ) -> TableNode:
        column_order = self._columns

        derived_header: RowNode | None = None
        if isinstance(rows, Mapping):
            row_nodes, derived_header = _rows_from_dict_of_lists(
                rows,
                header_styles=self._header_styles,
                column_order=column_order,
            )
        else:
            tupled_rows = _as_tuple(rows)
            if tupled_rows and all(isinstance(item, Mapping) for item in tupled_rows):
                typed_records = cast(
                    tuple[Mapping[ColumnKey, CellSource], ...], tupled_rows
                )
                row_nodes, derived_header = _rows_from_records(
                    typed_records,
                    header_styles=self._header_styles,
                    column_order=column_order,
                )
            else:
                row_nodes = tuple(_coerce_row(row) for row in tupled_rows)

        return TableNode(rows=row_nodes, styles=self._styles, header=derived_header)


class SheetBuilder:
    def __init__(
        self,
        name: str,
        *,
        background_color: str | None = None,
        show_gridlines: bool = True,
    ) -> None:
        self._name = name
        self._background_color = (
            normalize_hex(background_color) if background_color else None
        )
        self._show_gridlines = show_gridlines

    def __getitem__(
        self, items: SheetComponent | Sequence[SheetComponent]
    ) -> SheetNode:
        entries: list[SheetItem] = []
        for item in _as_tuple(items):
            if isinstance(item, Node):
                entries.append(item)
            else:
                msg = (
                    "Sheets accept rows, columns, tables, spacers, or layout stacks. "
                    "Call the builder before nesting."
                )
                raise TypeError(msg)
        return SheetNode(
            name=self._name,
            items=tuple(entries),
            background_color=self._background_color,
            show_gridlines=self._show_gridlines,
        )


class WorkbookBuilder:
    def __getitem__(
        self,
        sheets: SheetNode | ImportedSheetNode | Sequence[SheetNode | ImportedSheetNode],
    ) -> Workbook:
        sheet_nodes: list[SheetNode | ImportedSheetNode] = []
        for item in _as_tuple(sheets):
            if isinstance(item, (SheetNode, ImportedSheetNode)):
                sheet_nodes.append(item)
            else:
                raise TypeError(
                    "Workbooks accept sheet builders that have been indexed"
                )
        node = WorkbookNode(sheets=tuple(sheet_nodes))
        return Workbook(node)


def cell(
    *,
    style: Sequence[Style] | None = None,
    colspan: int = 1,
    rowspan: int = 1,
) -> CellBuilder:
    return CellBuilder(styles=style, colspan=colspan, rowspan=rowspan)


def row(*, style: Sequence[Style] | None = None) -> RowBuilder:
    return RowBuilder(styles=style)


def col(*, style: Sequence[Style] | None = None) -> ColumnBuilder:
    return ColumnBuilder(styles=style)


def table(
    *,
    style: Sequence[Style] | None = None,
    header_style: Sequence[Style] | None = None,
    column_order: Sequence[ColumnKey] | None = None,
) -> TableBuilder:
    return TableBuilder(styles=style, header_style=header_style, columns=column_order)


def sheet(
    name: str,
    *,
    background_color: str | None = None,
    show_gridlines: bool = True,
) -> SheetBuilder:
    return SheetBuilder(
        name,
        background_color=background_color,
        show_gridlines=show_gridlines,
    )


def import_sheet(
    source: str | Path | bytes | BinaryIO,
    sheet_name: str,
    *,
    name: str | None = None,
    show_gridlines: bool | None = None,
) -> ImportedSheetNode:
    """Import an existing sheet from a workbook without translating it."""

    dest_name = name or sheet_name
    return ImportedSheetNode(
        name=dest_name,
        source=source,
        source_sheet=sheet_name,
        show_gridlines=show_gridlines,
    )


def space(rows: int = 1, *, height: float | None = None) -> SpacerNode:
    if rows < 1:
        msg = "Spacer rows must be >= 1"
        raise ValueError(msg)
    return SpacerNode(rows=rows, height=height)


def vstack(
    *items: SheetComponent, gap: int = 0, style: Sequence[Style] | None = None
) -> VerticalStackNode:
    if not items:
        raise ValueError("Vertical stack requires at least one item")
    if gap < 0:
        raise ValueError("Vertical stack gap must be >= 0")
    components = tuple(_ensure_component(item) for item in items)
    return VerticalStackNode(items=components, gap=gap, styles=tuple(style or ()))


def hstack(
    *items: SheetComponent, gap: int = 0, style: Sequence[Style] | None = None
) -> HorizontalStackNode:
    if not items:
        raise ValueError("Horizontal stack requires at least one item")
    if gap < 0:
        raise ValueError("Horizontal stack gap must be >= 0")
    components = tuple(_ensure_component(item) for item in items)
    return HorizontalStackNode(items=components, gap=gap, styles=tuple(style or ()))


def workbook() -> WorkbookBuilder:
    return WorkbookBuilder()



"""HTML rendering engine implementation."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import date, datetime, time
from html import escape
from io import BytesIO
from pathlib import Path
from typing import BinaryIO

from openpyxl import load_workbook
from openpyxl.cell.cell import MergedCell

from ..styles import BorderStyleName, normalize_hex
from .base import EffectiveStyle, Engine, SaveTarget

__all__ = ["HtmlEngine"]

DEFAULT_FONT_NAME = "Calibri"
DEFAULT_FONT_SIZE = 11.0
DEFAULT_TEXT_COLOR = normalize_hex("#000000")
DEFAULT_BORDER_COLOR = normalize_hex("#000000")

COLUMN_WIDTH_PX = 7.0
BASE_CELL_PADDING_PX = 6


@dataclass
class _CellData:
    value: object
    style: EffectiveStyle
    border_fallback_color: str


@dataclass
class _SheetData:
    name: str
    show_gridlines: bool = True
    cells: dict[tuple[int, int], _CellData] = field(default_factory=dict)
    merges: dict[tuple[int, int], tuple[int, int]] = field(default_factory=dict)
    covered_cells: set[tuple[int, int]] = field(default_factory=set)
    col_widths: dict[int, float] = field(default_factory=dict)
    row_heights: dict[int, float] = field(default_factory=dict)
    background_color: str | None = None
    background_max_row: int = 0
    background_max_col: int = 0
    max_row: int = 0
    max_col: int = 0


class HtmlEngine(Engine):
    """Rendering engine that outputs HTML with Tailwind CSS."""

    def __init__(self) -> None:
        super().__init__()
        self._sheets: list[_SheetData] = []
        self._current_sheet: _SheetData | None = None

    def create_sheet(self, name: str, show_gridlines: bool = True) -> None:
        sheet = _SheetData(name=name, show_gridlines=show_gridlines)
        self._sheets.append(sheet)
        self._current_sheet = sheet

    def write_cell(
        self,
        row: int,
        col: int,
        value: object,
        style: EffectiveStyle,
        border_fallback_color: str,
    ) -> None:
        sheet = self._require_sheet()
        sheet.cells[(row, col)] = _CellData(
            value=value,
            style=style,
            border_fallback_color=normalize_hex(border_fallback_color),
        )
        sheet.max_row = max(sheet.max_row, row)
        sheet.max_col = max(sheet.max_col, col)

    def write_merged_cell(
        self,
        row: int,
        col: int,
        rowspan: int,
        colspan: int,
        value: object,
        style: EffectiveStyle,
        border_fallback_color: str,
    ) -> None:
        sheet = self._require_sheet()
        sheet.cells[(row, col)] = _CellData(
            value=value,
            style=style,
            border_fallback_color=normalize_hex(border_fallback_color),
        )
        sheet.merges[(row, col)] = (rowspan, colspan)
        for row_idx in range(row, row + rowspan):
            for col_idx in range(col, col + colspan):
                if row_idx == row and col_idx == col:
                    continue
                sheet.covered_cells.add((row_idx, col_idx))
        sheet.max_row = max(sheet.max_row, row + rowspan - 1)
        sheet.max_col = max(sheet.max_col, col + colspan - 1)

    def set_column_width(self, col: int, width: float) -> None:
        sheet = self._require_sheet()
        sheet.col_widths[col] = width
        sheet.max_col = max(sheet.max_col, col)

    def set_row_height(self, row: int, height: float) -> None:
        sheet = self._require_sheet()
        sheet.row_heights[row] = height
        sheet.max_row = max(sheet.max_row, row)

    def fill_background(self, color: str, max_row: int, max_col: int) -> None:
        sheet = self._require_sheet()
        sheet.background_color = normalize_hex(color)
        sheet.background_max_row = max(sheet.background_max_row, max_row)
        sheet.background_max_col = max(sheet.background_max_col, max_col)

    def copy_sheet(
        self,
        source: SaveTarget | bytes | BinaryIO,
        sheet_name: str,
        dest_name: str,
        show_gridlines: bool | None = None,
    ) -> None:
        source_wb = self._load_source_workbook(source)
        if sheet_name not in source_wb.sheetnames:
            raise ValueError(f"Sheet '{sheet_name}' not found in source workbook")
        source_ws = source_wb[sheet_name]

        source_gridlines = source_ws.sheet_view.showGridLines
        self.create_sheet(
            dest_name,
            show_gridlines=(
                True if source_gridlines is None else source_gridlines
            )
            if show_gridlines is None
            else show_gridlines,
        )
        sheet = self._require_sheet()
        merge_anchors: set[tuple[int, int]] = set()

        for merged_range in source_ws.merged_cells.ranges:
            min_col, min_row, max_col, max_row = merged_range.bounds
            sheet.merges[(min_row, min_col)] = (
                max_row - min_row + 1,
                max_col - min_col + 1,
            )
            merge_anchors.add((min_row, min_col))
            for row_idx in range(min_row, max_row + 1):
                for col_idx in range(min_col, max_col + 1):
                    if row_idx == min_row and col_idx == min_col:
                        continue
                    sheet.covered_cells.add((row_idx, col_idx))
            sheet.max_row = max(sheet.max_row, max_row)
            sheet.max_col = max(sheet.max_col, max_col)

        for row in source_ws.iter_rows():
            for cell in row:
                if isinstance(cell, MergedCell):
                    continue
                if cell.value is None and not getattr(cell, "has_style", False):
                    if (cell.row, cell.column) not in merge_anchors:
                        continue
                if cell.row is None or cell.column is None:
                    continue
                effective = self._openpyxl_cell_to_effective_style(cell)
                sheet.cells[(cell.row, cell.column)] = _CellData(
                    value=cell.value,
                    style=effective,
                    border_fallback_color=DEFAULT_BORDER_COLOR,
                )
                sheet.max_row = max(sheet.max_row, cell.row)
                sheet.max_col = max(sheet.max_col, cell.column)

        for col_letter, dim in source_ws.column_dimensions.items():
            if dim.width:
                col_idx = _column_letter_to_index(col_letter)
                sheet.col_widths[col_idx] = dim.width
                sheet.max_col = max(sheet.max_col, col_idx)

        for row_idx, dim in source_ws.row_dimensions.items():
            if dim.height:
                sheet.row_heights[row_idx] = dim.height
                sheet.max_row = max(sheet.max_row, row_idx)

    def save(self, target: SaveTarget | None = None) -> bytes | None:
        html = self._render_html()
        html_bytes = html.encode("utf-8")

        if target is None:
            return html_bytes
        if isinstance(target, (str, Path)):
            Path(target).write_bytes(html_bytes)
        else:
            target.write(html_bytes)
            if hasattr(target, "flush"):
                target.flush()
        return None

    def _render_html(self) -> str:
        tabs_html: list[str] = []
        sheets_html: list[str] = []

        for idx, sheet in enumerate(self._sheets):
            sheet_id = f"sheet-{idx}"
            is_active = idx == 0

            base_tab_classes = " ".join(
                [
                    "px-4",
                    "py-2",
                    "border",
                    "border-b-0",
                    "rounded-t",
                    "text-sm",
                    "font-medium",
                    "transition",
                    "duration-150",
                ]
            )
            active_tab_classes = "bg-white text-gray-900 border-gray-300"
            inactive_tab_classes = "bg-gray-200 text-gray-600 border-gray-200"
            tab_classes = (
                f"{base_tab_classes} "
                f"{active_tab_classes if is_active else inactive_tab_classes}"
            )
            tabs_html.append(
                """
                <button
                    type="button"
                    class="{tab_classes}"
                    data-sheet="{sheet_id}"
                >{label}</button>
                """.format(
                    tab_classes=tab_classes,
                    sheet_id=sheet_id,
                    label=escape(sheet.name),
                ).strip()
            )

            display = "block" if is_active else "none"
            bg_style = (
                f"background-color: {sheet.background_color};"
                if sheet.background_color
                else ""
            )
            table_html = self._render_sheet_table(sheet)
            sheets_html.append(
                """
                <div
                    id="{sheet_id}"
                    class="sheet-content"
                    style="display: {display}; {bg_style}"
                >{table}</div>
                """.format(
                    sheet_id=sheet_id,
                    display=display,
                    bg_style=bg_style,
                    table=table_html,
                ).strip()
            )

        tabs_markup = "".join(tabs_html)
        sheets_markup = "".join(sheets_html) if sheets_html else ""
        return f"""<!DOCTYPE html>
<html lang=\"en\">
<head>
  <meta charset=\"UTF-8\">
  <meta name=\"viewport\" content=\"width=device-width, initial-scale=1.0\">
  <title>Workbook</title>
  <script src=\"https://cdn.tailwindcss.com\"></script>
  <style>
    .sheet-content {{ min-height: 320px; overflow: auto; padding: 16px; }}
    table {{ border-collapse: collapse; }}
    td {{ padding: 2px 6px; vertical-align: top; }}
    .sheet-gridlines td {{ border: 1px solid #E5E7EB; }}
  </style>
</head>
<body class=\"bg-slate-100 p-6\">
  <div class=\"max-w-full\">
    <div class=\"border border-gray-300 rounded-lg bg-white shadow-sm\">
      <div class=\"flex gap-2 border-b border-gray-300 px-3 pt-3\">
        {tabs_markup}
      </div>
      <div class=\"p-2\">
        {sheets_markup}
      </div>
    </div>
  </div>
  <script>
    const tabs = Array.from(document.querySelectorAll(\"[data-sheet]\"));
    const sheets = Array.from(document.querySelectorAll(\".sheet-content\"));
    const activeClasses = [\"bg-white\", \"text-gray-900\", \"border-gray-300\"];
    const inactiveClasses = [\"bg-gray-200\", \"text-gray-600\", \"border-gray-200\"];

    const setActive = (sheetId, activeButton) => {{
      sheets.forEach((panel) => {{
        panel.style.display = panel.id === sheetId ? \"block\" : \"none\";
      }});
      tabs.forEach((button) => {{
        const isActive = button === activeButton;
        button.classList.remove(...activeClasses, ...inactiveClasses);
        button.classList.add(...(isActive ? activeClasses : inactiveClasses));
      }});
    }};

    if (tabs.length > 0) {{
      tabs.forEach((button) => {{
        button.addEventListener(\"click\", () => {{
          setActive(button.dataset.sheet, button);
        }});
      }});
      setActive(tabs[0].dataset.sheet, tabs[0]);
    }}
  </script>
</body>
</html>
"""

    def _render_sheet_table(self, sheet: _SheetData) -> str:
        max_row = max(sheet.max_row, sheet.background_max_row)
        max_col = max(sheet.max_col, sheet.background_max_col)
        max_row = max(max_row, max(sheet.row_heights.keys(), default=0))
        max_col = max(max_col, max(sheet.col_widths.keys(), default=0))

        if max_row == 0 or max_col == 0:
            return "<div class=\"text-gray-500\">Empty sheet</div>"

        colgroup = self._render_colgroup(max_col, sheet.col_widths)
        rows_html: list[str] = []
        for row in range(1, max_row + 1):
            cells_html: list[str] = []
            row_height = sheet.row_heights.get(row)
            for col in range(1, max_col + 1):
                if (row, col) in sheet.covered_cells:
                    continue
                cell_data = sheet.cells.get((row, col))
                rowspan, colspan = sheet.merges.get((row, col), (1, 1))
                if cell_data:
                    classes, inline_css = self._style_to_css(
                        cell_data.style, cell_data.border_fallback_color
                    )
                    value = _format_value(cell_data.value, cell_data.style.number_format)
                else:
                    classes = ""
                    inline_css = ""
                    value = ""
                span_attrs: list[str] = []
                if rowspan > 1:
                    span_attrs.append(f'rowspan="{rowspan}"')
                if colspan > 1:
                    span_attrs.append(f'colspan="{colspan}"')
                span_markup = f" {' '.join(span_attrs)}" if span_attrs else ""
                cells_html.append(
                    "<td class=\"{classes}\" style=\"{style}\"{span}>{value}</td>".format(
                        classes=classes,
                        style=inline_css,
                        span=span_markup,
                        value=escape(value),
                    )
                )
            row_style = f"height: {row_height}pt;" if row_height else ""
            rows_html.append(
                "<tr style=\"{style}\">{cells}</tr>".format(
                    style=row_style, cells="".join(cells_html)
                )
            )

        table_classes = "text-sm"
        if sheet.show_gridlines:
            table_classes += " sheet-gridlines"
        return "<table class=\"{classes}\">{colgroup}{rows}</table>".format(
            classes=table_classes,
            colgroup=colgroup,
            rows="".join(rows_html),
        )

    def _render_colgroup(self, max_col: int, col_widths: dict[int, float]) -> str:
        if not col_widths:
            return ""
        cols: list[str] = []
        for col in range(1, max_col + 1):
            width = col_widths.get(col)
            if width:
                px_width = width * COLUMN_WIDTH_PX
                cols.append(f"<col style=\"width: {px_width:.1f}px;\">")
            else:
                cols.append("<col>")
        return "<colgroup>{}</colgroup>".format("".join(cols))

    def _style_to_css(
        self, style: EffectiveStyle, border_fallback_color: str
    ) -> tuple[str, str]:
        classes: list[str] = []
        css_parts: list[str] = []

        css_parts.append("font-family: '{}' , sans-serif".format(style.font_name))
        css_parts.append("font-size: {}pt".format(style.font_size))
        css_parts.append("color: {}".format(style.text_color))

        if style.fill_color:
            css_parts.append("background-color: {}".format(style.fill_color))

        if style.bold:
            classes.append("font-bold")
        if style.italic:
            classes.append("italic")

        align_map = {
            "left": "text-left",
            "center": "text-center",
            "right": "text-right",
        }
        if style.horizontal_align in align_map:
            classes.append(align_map[style.horizontal_align])

        valign_map = {
            "top": "align-top",
            "center": "align-middle",
            "bottom": "align-bottom",
        }
        if style.vertical_align in valign_map:
            classes.append(valign_map[style.vertical_align])

        if style.indent:
            padding = BASE_CELL_PADDING_PX + style.indent * 8
            css_parts.append(f"padding-left: {padding}px")

        if style.wrap_text:
            classes.append("whitespace-normal")
            classes.append("break-words")
        else:
            classes.append("whitespace-nowrap")

        if style.shrink_to_fit:
            classes.append("truncate")

        border_css = self._border_to_css(style, border_fallback_color)
        if border_css:
            css_parts.extend(border_css)

        return " ".join(classes).strip(), "; ".join(css_parts).strip()

    def _border_to_css(
        self, style: EffectiveStyle, border_fallback_color: str
    ) -> list[str]:
        if style.border == "none":
            return []
        if not style.border and not (
            style.border_top
            or style.border_bottom
            or style.border_left
            or style.border_right
        ):
            return []

        border_style: BorderStyleName = style.border or "thin"
        border_color = style.border_color or border_fallback_color
        width = _border_style_to_width(border_style)
        css_style = _border_style_to_css(border_style)

        parts: list[str] = []
        if style.border_top or style.border_bottom or style.border_left or style.border_right:
            if style.border_top:
                parts.append(f"border-top: {width} {css_style} {border_color}")
            if style.border_bottom:
                parts.append(f"border-bottom: {width} {css_style} {border_color}")
            if style.border_left:
                parts.append(f"border-left: {width} {css_style} {border_color}")
            if style.border_right:
                parts.append(f"border-right: {width} {css_style} {border_color}")
        else:
            parts.append(f"border: {width} {css_style} {border_color}")
        return parts

    def _require_sheet(self) -> _SheetData:
        if self._current_sheet is None:
            raise RuntimeError("No sheet created. Call create_sheet() first.")
        return self._current_sheet

    def _load_source_workbook(self, source: SaveTarget | bytes | BinaryIO):
        if isinstance(source, (str, Path)):
            return load_workbook(filename=source, data_only=False)
        if isinstance(source, bytes):
            buffer: BinaryIO = BytesIO(source)
        else:
            buffer = source
            if hasattr(buffer, "seek"):
                try:
                    buffer.seek(0)
                except Exception:
                    pass
        return load_workbook(buffer, data_only=False)

    def _openpyxl_cell_to_effective_style(self, cell) -> EffectiveStyle:
        font = cell.font
        fill = cell.fill
        alignment = cell.alignment
        border = cell.border

        font_name = font.name if font and font.name else DEFAULT_FONT_NAME
        font_size = float(font.size) if font and font.size else DEFAULT_FONT_SIZE
        bold = bool(font.bold) if font else False
        italic = bool(font.italic) if font else False

        text_color = _openpyxl_color_to_hex(font.color) or DEFAULT_TEXT_COLOR

        fill_color = None
        if fill and getattr(fill, "fill_type", None) == "solid":
            fill_color = _openpyxl_color_to_hex(getattr(fill, "fgColor", None))

        horizontal_align = alignment.horizontal if alignment else None
        vertical_align = alignment.vertical if alignment else None
        indent = alignment.indent if alignment and alignment.indent else None
        wrap_text = bool(alignment.wrap_text) if alignment else False
        shrink_to_fit = bool(alignment.shrink_to_fit) if alignment else False

        border_style = None
        border_color = None
        border_top = False
        border_bottom = False
        border_left = False
        border_right = False

        for side_name in ("top", "bottom", "left", "right"):
            side = getattr(border, side_name, None)
            if not side or not side.style or side.style == "none":
                continue
            if border_style is None:
                border_style = side.style
            if side.color and not border_color:
                border_color = _openpyxl_color_to_hex(side.color)
            if side_name == "top":
                border_top = True
            elif side_name == "bottom":
                border_bottom = True
            elif side_name == "left":
                border_left = True
            elif side_name == "right":
                border_right = True

        return EffectiveStyle(
            font_name=font_name,
            font_size=font_size,
            bold=bold,
            italic=italic,
            text_color=text_color,
            fill_color=fill_color,
            horizontal_align=horizontal_align,
            vertical_align=vertical_align,
            indent=indent,
            wrap_text=wrap_text,
            shrink_to_fit=shrink_to_fit,
            auto_width=True,
            row_height=None,
            row_width=None,
            number_format=cell.number_format if cell.number_format != "General" else None,
            border=border_style,
            border_color=border_color,
            border_top=border_top,
            border_bottom=border_bottom,
            border_left=border_left,
            border_right=border_right,
        )


def _border_style_to_width(border: BorderStyleName) -> str:
    width_map = {
        "thin": "1px",
        "hair": "1px",
        "medium": "2px",
        "mediumDashed": "2px",
        "mediumDashDot": "2px",
        "mediumDashDotDot": "2px",
        "thick": "3px",
        "double": "3px",
    }
    return width_map.get(border, "1px")


def _border_style_to_css(border: BorderStyleName) -> str:
    style_map = {
        "thin": "solid",
        "medium": "solid",
        "thick": "solid",
        "hair": "solid",
        "dashed": "dashed",
        "mediumDashed": "dashed",
        "dotted": "dotted",
        "double": "double",
        "dashDot": "dashed",
        "mediumDashDot": "dashed",
        "dashDotDot": "dotted",
        "mediumDashDotDot": "dotted",
        "slantDashDot": "dashed",
    }
    return style_map.get(border, "solid")


def _openpyxl_color_to_hex(color, fallback: str | None = None) -> str | None:
    if color is None:
        return fallback
    rgb = getattr(color, "rgb", None)
    if isinstance(rgb, str) and len(rgb) >= 6:
        return normalize_hex("#" + rgb[-6:])
    return fallback


def _format_value(value: object, number_format: str | None) -> str:
    if value is None:
        return ""
    if isinstance(value, bool):
        return "TRUE" if value else "FALSE"
    if isinstance(value, (date, datetime, time)):
        if isinstance(value, datetime):
            return value.isoformat(sep=" ")
        return value.isoformat()
    if number_format and isinstance(value, (int, float)):
        if "%" in number_format:
            return f"{value * 100:.2f}%"
        if "$" in number_format:
            return f"${value:,.2f}"
        if "€" in number_format:
            return f"€{value:,.2f}"
        if "#,##0" in number_format:
            if ".00" in number_format or "0.00" in number_format:
                return f"{value:,.2f}"
            return f"{value:,.0f}"
    return str(value)


def _column_letter_to_index(letter: str) -> int:
    result = 0
    for char in letter.upper():
        result = result * 26 + (ord(char) - ord("A") + 1)
    return result



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import BinaryIO, TypeAlias

from .styles import Style

__all__ = [
    "CellNode",
    "RowNode",
    "ColumnNode",
    "TableNode",
    "SpacerNode",
    "VerticalStackNode",
    "HorizontalStackNode",
    "SheetNode",
    "ImportedSheetNode",
    "WorkbookNode",
    "SheetItem",
    "RenderableItem",
    "CellValue",
]

CellValue: TypeAlias = object


@dataclass(frozen=True)
class CellNode:
    value: CellValue
    styles: tuple[Style, ...] = ()
    colspan: int = 1
    rowspan: int = 1


@dataclass(frozen=True)
class RowNode:
    cells: tuple[CellNode, ...]
    styles: tuple[Style, ...] = ()


@dataclass(frozen=True)
class ColumnNode:
    cells: tuple[CellNode, ...]
    styles: tuple[Style, ...] = ()


@dataclass(frozen=True)
class TableNode:
    rows: tuple[RowNode, ...]
    styles: tuple[Style, ...] = ()
    header: RowNode | None = None


@dataclass(frozen=True)
class SpacerNode:
    rows: int = 1
    height: float | None = None


@dataclass(frozen=True)
class VerticalStackNode:
    items: tuple["SheetComponent", ...]
    gap: int = 0
    styles: tuple[Style, ...] = ()


@dataclass(frozen=True)
class HorizontalStackNode:
    items: tuple["SheetComponent", ...]
    gap: int = 0
    styles: tuple[Style, ...] = ()


SheetComponent = (
    CellNode
    | RowNode
    | ColumnNode
    | TableNode
    | SpacerNode
    | VerticalStackNode
    | HorizontalStackNode
)


RenderableItem = CellNode | RowNode | ColumnNode | TableNode | SpacerNode


SheetItem = SheetComponent


@dataclass(frozen=True)
class SheetNode:
    name: str
    items: tuple[SheetItem, ...]
    background_color: str | None = None
    show_gridlines: bool = True


@dataclass(frozen=True)
class ImportedSheetNode:
    """Reference to an existing Excel sheet to be copied as-is."""

    name: str
    source: str | Path | bytes | BinaryIO
    source_sheet: str
    show_gridlines: bool | None = None


@dataclass(frozen=True)
class WorkbookNode:
    sheets: tuple[SheetNode | ImportedSheetNode, ...]



from __future__ import annotations

import math
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import Literal, assert_never

from .engines.base import EffectiveStyle, Engine
from .nodes import (
    CellNode,
    ColumnNode,
    HorizontalStackNode,
    ImportedSheetNode,
    RowNode,
    SheetComponent,
    SheetNode,
    SpacerNode,
    TableNode,
    VerticalStackNode,
)
from .styles import (
    DEFAULT_BORDER_STYLE_NAME,
    BorderStyleName,
    Style,
    align_middle,
    bold,
    combine_styles,
    normalize_hex,
    text_center,
)

__all__ = ["render_sheet"]


DEFAULT_FONT_NAME = "Calibri"
DEFAULT_FONT_SIZE = 11.0
DEFAULT_MONO_FONT = "Consolas"
DEFAULT_TEXT_COLOR = normalize_hex("#000000")
DEFAULT_BORDER_COLOR = normalize_hex("#000000")
DEFAULT_BORDER_STYLE: BorderStyleName = DEFAULT_BORDER_STYLE_NAME
DEFAULT_ROW_HEIGHT = 16.0
DEFAULT_TABLE_HEADER_BG = None
DEFAULT_TABLE_HEADER_TEXT = None
DEFAULT_TABLE_STRIPE_COLOR = normalize_hex("#F2F4F7")
DEFAULT_TABLE_COMPACT_HEIGHT = 18.0
DEFAULT_BACKGROUND_MIN_ROWS = 200
DEFAULT_BACKGROUND_MIN_COLS = 80

_Axis = Literal["vertical", "horizontal"]


@dataclass(frozen=True)
class _PlacedCell:
    row: int
    col: int
    value: object
    styles: tuple[Style, ...]
    colspan: int
    rowspan: int
    prefer_height: float | None = None


@dataclass(frozen=True)
class _PlacedSpacer:
    row: int
    col: int
    rows: int
    height: float | None
    direction: _Axis


class _GridPlan:
    __slots__ = ("cells", "spacers", "_occupied", "max_row", "max_col")

    def __init__(self) -> None:
        self.cells: list[_PlacedCell] = []
        self.spacers: list[_PlacedSpacer] = []
        self._occupied: set[tuple[int, int]] = set()
        self.max_row = 0
        self.max_col = 0

    def is_clear(self, row: int, col: int, rowspan: int, colspan: int) -> bool:
        for row_idx in range(row, row + rowspan):
            for col_idx in range(col, col + colspan):
                if (row_idx, col_idx) in self._occupied:
                    return False
        return True

    def add_cell(
        self,
        row: int,
        col: int,
        value: object,
        styles: tuple[Style, ...],
        *,
        colspan: int = 1,
        rowspan: int = 1,
        prefer_height: float | None = None,
    ) -> None:
        if not self.is_clear(row, col, rowspan, colspan):
            msg = "Merged cells cannot overlap existing content"
            raise ValueError(msg)

        self.cells.append(
            _PlacedCell(
                row=row,
                col=col,
                value=value,
                styles=styles,
                colspan=colspan,
                rowspan=rowspan,
                prefer_height=prefer_height,
            )
        )
        for row_idx in range(row, row + rowspan):
            for col_idx in range(col, col + colspan):
                self._occupied.add((row_idx, col_idx))
        self.max_row = max(self.max_row, row + rowspan - 1)
        self.max_col = max(self.max_col, col + colspan - 1)

    def add_spacer(
        self,
        row: int,
        col: int,
        *,
        rows: int,
        height: float | None,
        direction: _Axis,
    ) -> None:
        self.spacers.append(
            _PlacedSpacer(
                row=row,
                col=col,
                rows=rows,
                height=height,
                direction=direction,
            )
        )
        if direction == "horizontal":
            self.max_row = max(self.max_row, row)
            self.max_col = max(self.max_col, col + rows - 1)
        else:
            self.max_row = max(self.max_row, row + rows - 1)
            self.max_col = max(self.max_col, col)

    def merge(self, other: _GridPlan, *, row: int, col: int) -> None:
        for placement in other.cells:
            self.add_cell(
                row + placement.row - 1,
                col + placement.col - 1,
                placement.value,
                placement.styles,
                colspan=placement.colspan,
                rowspan=placement.rowspan,
                prefer_height=placement.prefer_height,
            )
        for spacer in other.spacers:
            self.add_spacer(
                row + spacer.row - 1,
                col + spacer.col - 1,
                rows=spacer.rows,
                height=spacer.height,
                direction=spacer.direction,
            )
        if other.max_row > 0:
            self.max_row = max(self.max_row, row + other.max_row - 1)
        if other.max_col > 0:
            self.max_col = max(self.max_col, col + other.max_col - 1)


def _resolve(styles: Sequence[Style]) -> EffectiveStyle:
    base_style = Style(
        font_name=DEFAULT_FONT_NAME,
        font_size=DEFAULT_FONT_SIZE,
        text_color=DEFAULT_TEXT_COLOR,
    )
    merged = combine_styles(styles, base=base_style)

    font_name = merged.font_name or DEFAULT_FONT_NAME
    if merged.mono:
        font_name = DEFAULT_MONO_FONT
    font_size = merged.font_size if merged.font_size is not None else DEFAULT_FONT_SIZE
    if merged.font_size_delta is not None:
        font_size += merged.font_size_delta

    bold_flag = merged.bold if merged.bold is not None else False
    italic_flag = merged.italic if merged.italic is not None else False

    text_color = normalize_hex(merged.text_color or DEFAULT_TEXT_COLOR)
    fill_color = normalize_hex(merged.fill_color) if merged.fill_color else None
    border_color = normalize_hex(merged.border_color) if merged.border_color else None
    shrink_to_fit = merged.shrink_to_fit if merged.shrink_to_fit is not None else False
    auto_width = merged.auto_width if merged.auto_width is not None else True
    row_height = merged.row_height
    row_width = merged.row_width
    border_top = merged.border_top if merged.border_top is not None else False
    border_bottom = merged.border_bottom if merged.border_bottom is not None else False
    border_left = merged.border_left if merged.border_left is not None else False
    border_right = merged.border_right if merged.border_right is not None else False

    return EffectiveStyle(
        font_name=font_name,
        font_size=font_size,
        bold=bold_flag,
        italic=italic_flag,
        text_color=text_color,
        fill_color=fill_color,
        horizontal_align=merged.horizontal_align,
        vertical_align=merged.vertical_align,
        indent=merged.indent,
        wrap_text=merged.wrap_text if merged.wrap_text is not None else False,
        shrink_to_fit=shrink_to_fit,
        auto_width=auto_width,
        row_height=row_height,
        row_width=row_width,
        number_format=merged.number_format,
        border=merged.border,
        border_color=border_color,
        border_top=border_top,
        border_bottom=border_bottom,
        border_left=border_left,
        border_right=border_right,
    )


def _default_row_height() -> float:
    return DEFAULT_ROW_HEIGHT


def _estimate_wrap_lines(text: str) -> int:
    wrap_line_length = 30
    if not text:
        return 1
    lines = 0
    for raw_line in text.splitlines() or [text]:
        length = max(len(raw_line), 1)
        lines += max(1, math.ceil(length / wrap_line_length))
    return max(lines, 1)


def _update_dimensions(
    *,
    col_widths: dict[int, float],
    row_heights: dict[int, float],
    column_index: int,
    row_index: int,
    value: object,
    style: EffectiveStyle,
    colspan: int = 1,
    rowspan: int = 1,
    prefer_height: float | None = None,
) -> None:
    text = "" if value is None else str(value)
    font_scale = style.font_size / DEFAULT_FONT_SIZE if style.font_size else 1.0
    width_hint = max(len(text), 1.0)
    existing_total_width = sum(
        col_widths.get(col_idx, 0.0)
        for col_idx in range(column_index, column_index + colspan)
    )
    if style.row_width is not None:
        total_width = style.row_width
    elif not style.auto_width:
        total_width = existing_total_width if existing_total_width else 8.0 * colspan
    elif style.wrap_text:
        total_width = existing_total_width or 8.0 * colspan
    else:
        total_width = width_hint * font_scale + 1.0
    per_column_width = total_width / colspan
    for col_idx in range(column_index, column_index + colspan):
        col_widths[col_idx] = max(col_widths.get(col_idx, 0.0), per_column_width)

    if style.row_height is not None:
        per_row_height = style.row_height
    else:
        base_height = (
            prefer_height if prefer_height is not None else _default_row_height()
        )
        if style.wrap_text:
            base_height *= _estimate_wrap_lines(text)
        base_height *= font_scale
        base_height += 2.0
        per_row_height = base_height / rowspan
    for row_idx in range(row_index, row_index + rowspan):
        row_heights[row_idx] = max(row_heights.get(row_idx, 0.0), per_row_height)


def _find_next_clear_col(
    plan: _GridPlan,
    *,
    row: int,
    start_col: int,
    rowspan: int,
    colspan: int,
) -> int:
    candidate = max(start_col, 1)
    while not plan.is_clear(row, candidate, rowspan, colspan):
        candidate += 1
    return candidate


def _find_next_clear_row(
    plan: _GridPlan,
    *,
    start_row: int,
    col: int,
    rowspan: int,
    colspan: int,
) -> int:
    candidate = max(start_row, 1)
    while not plan.is_clear(candidate, col, rowspan, colspan):
        candidate += 1
    return candidate


def _place_row_node(
    plan: _GridPlan,
    node: RowNode,
    *,
    row: int,
    extra_styles: tuple[Style, ...],
) -> None:
    cursor = 1
    for cell_node in node.cells:
        column_index = _find_next_clear_col(
            plan,
            row=row,
            start_col=cursor,
            rowspan=cell_node.rowspan,
            colspan=cell_node.colspan,
        )
        plan.add_cell(
            row,
            column_index,
            cell_node.value,
            (*extra_styles, *node.styles, *cell_node.styles),
            colspan=cell_node.colspan,
            rowspan=cell_node.rowspan,
        )
        cursor = column_index + cell_node.colspan


def _place_column_node(
    plan: _GridPlan,
    node: ColumnNode,
    *,
    col: int,
    extra_styles: tuple[Style, ...],
) -> None:
    cursor = 1
    for cell_node in node.cells:
        row_index = _find_next_clear_row(
            plan,
            start_row=cursor,
            col=col,
            rowspan=cell_node.rowspan,
            colspan=cell_node.colspan,
        )
        plan.add_cell(
            row_index,
            col,
            cell_node.value,
            (*extra_styles, *node.styles, *cell_node.styles),
            colspan=cell_node.colspan,
            rowspan=cell_node.rowspan,
        )
        cursor = row_index + 1


def _place_single_cell(
    plan: _GridPlan,
    node: CellNode,
    *,
    row: int,
    extra_styles: tuple[Style, ...],
) -> None:
    column_index = _find_next_clear_col(
        plan,
        row=row,
        start_col=1,
        rowspan=node.rowspan,
        colspan=node.colspan,
    )
    plan.add_cell(
        row,
        column_index,
        node.value,
        (*extra_styles, *node.styles),
        colspan=node.colspan,
        rowspan=node.rowspan,
    )


def _table_has_merged_cells(node: TableNode) -> bool:
    if node.header and any(
        cell.colspan > 1 or cell.rowspan > 1 for cell in node.header.cells
    ):
        return True
    return any(
        cell.colspan > 1 or cell.rowspan > 1
        for row in node.rows
        for cell in row.cells
    )


def _build_table_plan(node: TableNode, extra_styles: tuple[Style, ...]) -> _GridPlan:
    if _table_has_merged_cells(node):
        msg = "Merged cells are not supported inside tables"
        raise ValueError(msg)

    plan = _GridPlan()
    table_style = combine_styles((*extra_styles, *node.styles))
    banded = table_style.table_banded if table_style.table_banded is not None else False
    bordered = (
        table_style.table_bordered if table_style.table_bordered is not None else True
    )
    compact = (
        table_style.table_compact if table_style.table_compact is not None else False
    )
    border_color = (
        table_style.border_color
        if table_style.border_color is not None
        else DEFAULT_BORDER_COLOR
    )
    border_style = (
        table_style.border if table_style.border is not None else DEFAULT_BORDER_STYLE
    )

    table_border_style = (
        Style(border=border_style, border_color=border_color) if bordered else None
    )
    stripe_style = Style(fill_color=DEFAULT_TABLE_STRIPE_COLOR) if banded else None
    compact_height = DEFAULT_TABLE_COMPACT_HEIGHT if compact else None

    current_row = 1

    def add_row(
        row_node: RowNode,
        *,
        extras: Sequence[Style] = (),
        prefer_height: float | None = None,
        extras_first: bool = False,
    ) -> None:
        for column_offset, cell_node in enumerate(row_node.cells, start=1):
            base_chain = (*extra_styles, *node.styles)
            if extras_first:
                style_chain = (*base_chain, *extras, *row_node.styles, *cell_node.styles)
            else:
                style_chain = (*base_chain, *row_node.styles, *extras, *cell_node.styles)
            if table_border_style:
                style_chain = (*style_chain, table_border_style)
            plan.add_cell(
                current_row,
                column_offset,
                cell_node.value,
                style_chain,
                prefer_height=prefer_height,
            )

    if node.header:
        header_extras: list[Style] = [bold, text_center, align_middle]
        if DEFAULT_TABLE_HEADER_BG:
            header_extras.append(Style(fill_color=DEFAULT_TABLE_HEADER_BG))
        if DEFAULT_TABLE_HEADER_TEXT:
            header_extras.append(Style(text_color=DEFAULT_TABLE_HEADER_TEXT))
        add_row(
            node.header,
            extras=header_extras,
            prefer_height=compact_height,
            extras_first=True,
        )
        current_row += 1

    for idx, row_node in enumerate(node.rows):
        extras: list[Style] = []
        if stripe_style and idx % 2 == 1:
            extras.append(stripe_style)
        add_row(row_node, extras=extras, prefer_height=compact_height)
        current_row += 1

    return plan


def _build_horizontal_plan(
    items: Sequence[SheetComponent],
    *,
    extra_styles: tuple[Style, ...],
    gap: int,
) -> _GridPlan:
    plan = _GridPlan()
    col_cursor = 1
    for idx, child in enumerate(items):
        if isinstance(child, SpacerNode):
            plan.add_spacer(
                1,
                col_cursor,
                rows=child.rows,
                height=child.height,
                direction="horizontal",
            )
            col_cursor += child.rows
        else:
            child_plan = _build_item_plan(child, extra_styles=extra_styles)
            plan.merge(child_plan, row=1, col=col_cursor)
            col_cursor += _logical_width(child)
        if idx < len(items) - 1:
            col_cursor += gap
    return plan


def _build_vertical_plan(
    items: Sequence[SheetComponent],
    *,
    extra_styles: tuple[Style, ...],
    gap: int,
) -> _GridPlan:
    plan = _GridPlan()
    row_cursor = 1
    for idx, child in enumerate(items):
        if isinstance(child, CellNode):
            _place_single_cell(plan, child, row=row_cursor, extra_styles=extra_styles)
            row_cursor += 1
        elif isinstance(child, RowNode):
            _place_row_node(plan, child, row=row_cursor, extra_styles=extra_styles)
            row_cursor += 1
        elif isinstance(child, SpacerNode):
            plan.add_spacer(
                row_cursor,
                1,
                rows=child.rows,
                height=child.height,
                direction="vertical",
            )
            row_cursor += child.rows
        else:
            child_plan = _build_item_plan(child, extra_styles=extra_styles)
            plan.merge(child_plan, row=row_cursor, col=1)
            row_cursor += _logical_height(child)
        if idx < len(items) - 1:
            row_cursor += gap
    return plan


def _logical_width(item: SheetComponent) -> int:
    if isinstance(item, CellNode):
        return 1
    if isinstance(item, RowNode):
        return len(item.cells)
    if isinstance(item, ColumnNode):
        return 1
    if isinstance(item, TableNode):
        width = 0
        if item.header:
            width = max(width, len(item.header.cells))
        for row in item.rows:
            width = max(width, len(row.cells))
        return width
    if isinstance(item, SpacerNode):
        return 1
    if isinstance(item, VerticalStackNode):
        return max(_logical_width(child) for child in item.items)
    if isinstance(item, HorizontalStackNode):
        total = sum(_logical_width(child) for child in item.items)
        total += item.gap * (len(item.items) - 1)
        return total
    assert_never(item)


def _logical_height(item: SheetComponent) -> int:
    if isinstance(item, CellNode):
        return 1
    if isinstance(item, RowNode):
        return 1
    if isinstance(item, ColumnNode):
        return len(item.cells)
    if isinstance(item, TableNode):
        return len(item.rows) + (1 if item.header else 0)
    if isinstance(item, SpacerNode):
        return item.rows
    if isinstance(item, VerticalStackNode):
        total = sum(_logical_height(child) for child in item.items)
        total += item.gap * (len(item.items) - 1)
        return total
    if isinstance(item, HorizontalStackNode):
        return max(_logical_height(child) for child in item.items)
    assert_never(item)


def _build_item_plan(
    item: SheetComponent,
    *,
    extra_styles: tuple[Style, ...] = (),
) -> _GridPlan:
    if isinstance(item, CellNode):
        plan = _GridPlan()
        plan.add_cell(
            1,
            1,
            item.value,
            (*extra_styles, *item.styles),
            colspan=item.colspan,
            rowspan=item.rowspan,
        )
        return plan
    if isinstance(item, RowNode):
        plan = _GridPlan()
        _place_row_node(plan, item, row=1, extra_styles=extra_styles)
        return plan
    if isinstance(item, ColumnNode):
        plan = _GridPlan()
        _place_column_node(plan, item, col=1, extra_styles=extra_styles)
        return plan
    if isinstance(item, TableNode):
        return _build_table_plan(item, extra_styles)
    if isinstance(item, SpacerNode):
        plan = _GridPlan()
        plan.add_spacer(
            1,
            1,
            rows=item.rows,
            height=item.height,
            direction="vertical",
        )
        return plan
    if isinstance(item, VerticalStackNode):
        return _build_vertical_plan(
            item.items,
            extra_styles=extra_styles + item.styles,
            gap=item.gap,
        )
    if isinstance(item, HorizontalStackNode):
        return _build_horizontal_plan(
            item.items,
            extra_styles=extra_styles + item.styles,
            gap=item.gap,
        )
    assert_never(item)


def _apply_dimensions(
    engine: Engine,
    col_widths: Mapping[int, float],
    row_heights: Mapping[int, float],
) -> None:
    for column_index, width in col_widths.items():
        engine.set_column_width(column_index, width)
    for row_index, height in row_heights.items():
        engine.set_row_height(row_index, height)


def render_sheet(engine: Engine, node: SheetNode | ImportedSheetNode) -> None:
    """Render a sheet node using the given engine."""
    if isinstance(node, ImportedSheetNode):
        engine.copy_sheet(
            node.source,
            node.source_sheet,
            node.name,
            show_gridlines=node.show_gridlines,
        )
        return

    engine.create_sheet(node.name, show_gridlines=node.show_gridlines)

    col_widths: dict[int, float] = {}
    row_heights: dict[int, float] = {}
    plan = _build_vertical_plan(node.items, extra_styles=(), gap=0)

    max_row = plan.max_row
    max_col = plan.max_col

    if node.background_color:
        normalized = normalize_hex(node.background_color)
        target_max_row = max(max_row, DEFAULT_BACKGROUND_MIN_ROWS)
        target_max_col = max(max_col, DEFAULT_BACKGROUND_MIN_COLS)
        engine.fill_background(normalized, target_max_row, target_max_col)

    for placement in sorted(plan.cells, key=lambda cell: (cell.row, cell.col)):
        effective = _resolve(placement.styles)
        if placement.colspan == 1 and placement.rowspan == 1:
            engine.write_cell(
                placement.row,
                placement.col,
                placement.value,
                effective,
                DEFAULT_BORDER_COLOR,
            )
        else:
            engine.write_merged_cell(
                placement.row,
                placement.col,
                placement.rowspan,
                placement.colspan,
                placement.value,
                effective,
                DEFAULT_BORDER_COLOR,
            )
        _update_dimensions(
            col_widths=col_widths,
            row_heights=row_heights,
            column_index=placement.col,
            row_index=placement.row,
            value=placement.value,
            style=effective,
            colspan=placement.colspan,
            rowspan=placement.rowspan,
            prefer_height=placement.prefer_height,
        )

    for spacer in plan.spacers:
        if spacer.direction == "horizontal":
            continue
        height = spacer.height if spacer.height is not None else _default_row_height()
        for offset in range(spacer.rows):
            row_index = spacer.row + offset
            row_heights[row_index] = max(row_heights.get(row_index, 0.0), height)

    _apply_dimensions(engine, col_widths, row_heights)



"""Tests for HTML rendering engine."""

from __future__ import annotations

import tempfile
from io import BytesIO
from pathlib import Path

import pytest

import xpyxl as x


def test_save_returns_bytes_when_target_none() -> None:
    workbook = x.workbook()[x.sheet("Test")[x.row()["Hello", "World"]]]
    result = workbook.save(engine="html")

    assert result is not None
    assert isinstance(result, bytes)
    assert b"<!DOCTYPE html>" in result
    assert b"Hello" in result
    assert b"World" in result


def test_save_writes_file_when_path_given() -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        output_path = Path(tmpdir) / "output.html"

        workbook = x.workbook()[x.sheet("Test")[x.row()["Data"]]]
        result = workbook.save(output_path, engine="html")

        assert result is None
        assert output_path.exists()
        content = output_path.read_text()
        assert "<!DOCTYPE html>" in content
        assert "Data" in content


def test_save_writes_to_stream() -> None:
    buffer = BytesIO()

    workbook = x.workbook()[x.sheet("Test")[x.row()["Stream"]]]
    result = workbook.save(buffer, engine="html")

    assert result is None
    content = buffer.getvalue().decode("utf-8")
    assert "<!DOCTYPE html>" in content
    assert "Stream" in content


def test_multiple_sheets_with_tabs() -> None:
    workbook = x.workbook()[
        x.sheet("Sheet1")[x.row()["A"]],
        x.sheet("Sheet2")[x.row()["B"]],
        x.sheet("Sheet3")[x.row()["C"]],
    ]
    result = workbook.save(engine="html")
    assert isinstance(result, bytes)
    html = result.decode("utf-8")

    assert "Sheet1" in html
    assert "Sheet2" in html
    assert "Sheet3" in html
    assert "data-sheet" in html
    assert html.count("class=\"sheet-content\"") == 3


def test_sheet_names_escaped() -> None:
    workbook = x.workbook()[x.sheet("<Script>")[x.row()["Safe"]]]
    result = workbook.save(engine="html")
    assert isinstance(result, bytes)
    html = result.decode("utf-8")

    assert "&lt;Script&gt;" in html
    assert "<Script>" not in html


def test_bold_italic_styles() -> None:
    workbook = x.workbook()[
        x.sheet("Test")[
            x.row()[x.cell(style=[x.bold])["Bold"]],
            x.row()[x.cell(style=[x.italic])["Italic"]],
        ]
    ]
    result = workbook.save(engine="html")
    assert isinstance(result, bytes)
    html = result.decode("utf-8")

    assert "font-bold" in html
    assert "italic" in html


def test_text_alignment() -> None:
    workbook = x.workbook()[
        x.sheet("Test")[
            x.row()[x.cell(style=[x.text_left])["Left"]],
            x.row()[x.cell(style=[x.text_center])["Center"]],
            x.row()[x.cell(style=[x.text_right])["Right"]],
        ]
    ]
    result = workbook.save(engine="html")
    assert isinstance(result, bytes)
    html = result.decode("utf-8")

    assert "text-left" in html
    assert "text-center" in html
    assert "text-right" in html


def test_background_color() -> None:
    workbook = x.workbook()[
        x.sheet("Test")[x.row()[x.cell(style=[x.bg_primary])["Colored"]]]
    ]
    result = workbook.save(engine="html")
    assert isinstance(result, bytes)
    html = result.decode("utf-8")

    assert "background-color:" in html


def test_borders() -> None:
    workbook = x.workbook()[
        x.sheet("Test")[x.row()[x.cell(style=[x.border_all])["Bordered"]]]
    ]
    result = workbook.save(engine="html")
    assert isinstance(result, bytes)
    html = result.decode("utf-8")

    assert "border:" in html or "border-top:" in html


def test_sheet_gridlines_can_be_hidden() -> None:
    workbook = x.workbook()[
        x.sheet("Test", show_gridlines=False)[x.row()["A", "B"]]
    ]
    result = workbook.save(engine="html")
    assert isinstance(result, bytes)
    html = result.decode("utf-8")

    assert 'class="text-sm sheet-gridlines"' not in html


def test_copy_sheet_from_xlsx() -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        source_path = Path(tmpdir) / "source.xlsx"
        source_wb = x.workbook()[
            x.sheet("Source")[x.row()["Imported", "Data"]]
        ]
        source_wb.save(source_path, engine="openpyxl")

        workbook = x.workbook()[x.import_sheet(source_path, "Source", name="Imported")]
        result = workbook.save(engine="html")
        assert isinstance(result, bytes)
        html = result.decode("utf-8")

        assert "Imported" in html
        assert "Data" in html


def test_copy_sheet_preserves_hidden_gridlines() -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        source_path = Path(tmpdir) / "source.xlsx"
        source_wb = x.workbook()[
            x.sheet("Source", show_gridlines=False)[x.row()["Imported", "Data"]]
        ]
        source_wb.save(source_path, engine="openpyxl")

        workbook = x.workbook()[x.import_sheet(source_path, "Source", name="Imported")]
        result = workbook.save(engine="html")
        assert isinstance(result, bytes)
        html = result.decode("utf-8")

        assert 'class="text-sm sheet-gridlines"' not in html


def test_copy_sheet_missing_sheet_raises() -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        source_path = Path(tmpdir) / "source.xlsx"
        source_wb = x.workbook()[x.sheet("Exists")[x.row()["Data"]]]
        source_wb.save(source_path, engine="openpyxl")

        workbook = x.workbook()[
            x.import_sheet(source_path, "NotExists", name="Imported")
        ]

        with pytest.raises(ValueError, match="not found"):
            workbook.save(engine="html")


def test_with_table() -> None:
    table = x.table()[
        [
            ["Name", "Value"],
            ["A", 1],
            ["B", 2],
        ]
    ]
    workbook = x.workbook()[x.sheet("Table")[table]]
    result = workbook.save(engine="html")
    assert isinstance(result, bytes)
    html = result.decode("utf-8")

    assert "Name" in html
    assert "Value" in html
    assert "<table" in html


def test_tailwind_cdn_included() -> None:
    workbook = x.workbook()[x.sheet("Test")[x.row()["Data"]]]
    result = workbook.save(engine="html")
    assert isinstance(result, bytes)
    html = result.decode("utf-8")

    assert "cdn.tailwindcss.com" in html



from __future__ import annotations

from pathlib import Path
from typing import BinaryIO

from openpyxl import Workbook as _OpenpyxlWorkbook

from .engines import EngineName, get_engine
from .nodes import WorkbookNode
from .render import render_sheet

__all__ = ["Workbook"]


class Workbook:
    """Immutable workbook aggregate with a `.save()` convenience."""

    def __init__(self, node: WorkbookNode) -> None:
        self._node = node

    def save(
        self,
        target: str | Path | BinaryIO | None = None,
        *,
        engine: EngineName = "hybrid",
    ) -> bytes | None:
        """Save the workbook to a file or binary stream.

        Args:
            target: File path or binary buffer to write to. Pass None to receive
                the rendered workbook as bytes.
            engine: The rendering engine to use. Options are:
                - "hybrid" (default): Combines xlsxwriter speed for generated sheets
                  with openpyxl for importing sheets. Best balance of speed and features.
                - "openpyxl": Full-featured engine supporting all features including
                  import_sheet.
                - "xlsxwriter": Fast generation engine. Does NOT support import_sheet;
                  use "hybrid" or "openpyxl" for that.
        """
        engine_instance = get_engine(engine)
        for sheet in self._node.sheets:
            render_sheet(engine_instance, sheet)
        return engine_instance.save(target)

    def to_openpyxl(self) -> _OpenpyxlWorkbook:
        """Convert to an openpyxl Workbook object.

        This method is provided for backward compatibility and advanced use cases
        where direct access to the openpyxl workbook is needed.
        """
        from .engines.openpyxl_engine import OpenpyxlEngine

        # Render with the openpyxl engine without persisting to disk
        engine = OpenpyxlEngine()
        for sheet in self._node.sheets:
            render_sheet(engine, sheet)
        return engine._workbook



"""Abstract base class for rendering engines."""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

if TYPE_CHECKING:
    from ..styles import BorderStyleName

__all__ = ["Engine", "EffectiveStyle", "SaveTarget"]


@dataclass
class EffectiveStyle:
    """Resolved style with all defaults applied."""

    font_name: str
    font_size: float
    bold: bool
    italic: bool
    text_color: str
    fill_color: str | None
    horizontal_align: str | None
    vertical_align: str | None
    indent: int | None
    wrap_text: bool
    shrink_to_fit: bool
    auto_width: bool
    row_height: float | None
    row_width: float | None
    number_format: str | None
    border: "BorderStyleName | None"
    border_color: str | None
    border_top: bool
    border_bottom: bool
    border_left: bool
    border_right: bool


SaveTarget = str | Path | BinaryIO


class Engine(ABC):
    """Abstract base class for Excel rendering engines.

    Engines handle the low-level details of writing to Excel files,
    including cell values, styles, and dimensions.

    Note on indexing:
    - The engine interface uses 1-based indexing for rows and columns
      (matching Excel conventions and openpyxl).
    - Implementations that use 0-based indexing (like xlsxwriter) must
      convert internally.
    """

    def __init__(self) -> None:
        """Initialize engine state without binding to an output target."""

    @abstractmethod
    def create_sheet(self, name: str, show_gridlines: bool = True) -> None:
        """Create a new worksheet with the given name."""
        ...

    @abstractmethod
    def write_cell(
        self,
        row: int,
        col: int,
        value: object,
        style: EffectiveStyle,
        border_fallback_color: str,
    ) -> None:
        """Write a value to a cell with the given style.

        Args:
            row: 1-based row index
            col: 1-based column index
            value: The cell value
            style: Resolved style to apply
            border_fallback_color: Default border color if not specified in style
        """
        ...

    @abstractmethod
    def write_merged_cell(
        self,
        row: int,
        col: int,
        rowspan: int,
        colspan: int,
        value: object,
        style: EffectiveStyle,
        border_fallback_color: str,
    ) -> None:
        """Write a merged cell anchored at the given row/column.

        Args:
            row: 1-based anchor row index
            col: 1-based anchor column index
            rowspan: Number of rows in the merged range
            colspan: Number of columns in the merged range
            value: The anchor cell value
            style: Resolved style to apply
            border_fallback_color: Default border color if not specified in style
        """
        ...

    @abstractmethod
    def set_column_width(self, col: int, width: float) -> None:
        """Set the width of a column.

        Args:
            col: 1-based column index
            width: Column width in characters
        """
        ...

    @abstractmethod
    def set_row_height(self, row: int, height: float) -> None:
        """Set the height of a row.

        Args:
            row: 1-based row index
            height: Row height in points
        """
        ...

    @abstractmethod
    def fill_background(
        self,
        color: str,
        max_row: int,
        max_col: int,
    ) -> None:
        """Fill the background of a range of cells.

        Args:
            color: Hex color code (e.g., "#FFFFFF")
            max_row: Maximum row to fill (1-based)
            max_col: Maximum column to fill (1-based)
        """
        ...

    @abstractmethod
    def copy_sheet(
        self,
        source: SaveTarget | bytes | BinaryIO,
        sheet_name: str,
        dest_name: str,
        show_gridlines: bool | None = None,
    ) -> None:
        """Copy an existing sheet from another workbook into this workbook.

        Args:
            source: Path, file-like, or bytes of the source workbook.
            sheet_name: Name of the sheet within the source workbook to copy.
            dest_name: Name of the sheet to create in the destination workbook.
            show_gridlines: Override gridline visibility. When None, preserve the
                source sheet setting if available.
        """
        ...

    @abstractmethod
    def save(self, target: SaveTarget | None = None) -> bytes | None:
        """Finalize and persist the workbook.

        Args:
            target: Destination to write the workbook to. If None, the engine
                should return the workbook as bytes. When provided, the engine
                writes to the given path or binary file-like object and should
                return None.
        """
        ...



from __future__ import annotations

from xpyxl.nodes import ImportedSheetNode, SheetNode

from ._workbook import Workbook
from .builders import (
    Node,
    cell,
    col,
    hstack,
    import_sheet,
    row,
    sheet,
    space,
    table,
    vstack,
    workbook,
)
from .styles import (
    BorderStyleLiteral,
    BorderStyleName,
    Style,
    align_bottom,
    align_middle,
    align_top,
    allow_overflow,
    bg_info,
    bg_muted,
    bg_primary,
    bg_red,
    bg_success,
    bg_warning,
    bold,
    border_all,
    border_black,
    border_blue,
    border_bottom,
    border_dashed,
    border_dotted,
    border_double,
    border_gray,
    border_green,
    border_left,
    border_medium,
    border_muted,
    border_none,
    border_orange,
    border_primary,
    border_purple,
    border_red,
    border_right,
    border_thick,
    border_thin,
    border_top,
    border_white,
    border_x,
    border_y,
    combine_styles,
    currency_eur,
    currency_usd,
    date_short,
    datetime_short,
    italic,
    mono,
    muted,
    normalize_hex,
    nowrap,
    number_comma,
    number_precision,
    percent,
    row_height,
    row_width,
    table_banded,
    table_bordered,
    table_compact,
    text_2xl,
    text_3xl,
    text_base,
    text_black,
    text_blue,
    text_center,
    text_gray,
    text_green,
    text_left,
    text_lg,
    text_muted,
    text_orange,
    text_primary,
    text_purple,
    text_red,
    text_right,
    text_sm,
    text_white,
    text_xl,
    text_xs,
    time_short,
    to_argb,
    wrap,
    wrap_shrink,
)

__all__ = [
    "Workbook",
    "SheetNode",
    "ImportedSheetNode",
    "Node",
    "Style",
    "BorderStyleName",
    "BorderStyleLiteral",
    "workbook",
    "sheet",
    "import_sheet",
    "row",
    "col",
    "cell",
    "table",
    "space",
    "vstack",
    "hstack",
    "combine_styles",
    "normalize_hex",
    "to_argb",
    "text_xs",
    "text_sm",
    "text_base",
    "text_lg",
    "text_xl",
    "text_2xl",
    "text_3xl",
    "bold",
    "italic",
    "mono",
    "muted",
    "text_muted",
    "text_primary",
    "text_white",
    "text_red",
    "text_green",
    "text_blue",
    "text_orange",
    "text_purple",
    "text_black",
    "text_gray",
    "text_left",
    "text_center",
    "text_right",
    "align_top",
    "align_middle",
    "align_bottom",
    "wrap",
    "nowrap",
    "wrap_shrink",
    "allow_overflow",
    "row_height",
    "row_width",
    "nowrap",
    "wrap_shrink",
    "border_all",
    "border_top",
    "border_bottom",
    "border_left",
    "border_right",
    "border_x",
    "border_y",
    "border_red",
    "border_green",
    "border_blue",
    "border_orange",
    "border_purple",
    "border_black",
    "border_gray",
    "border_white",
    "border_muted",
    "border_primary",
    "border_thin",
    "border_medium",
    "border_thick",
    "border_dashed",
    "border_dotted",
    "border_double",
    "border_none",
    "bg_red",
    "bg_primary",
    "bg_muted",
    "bg_success",
    "bg_warning",
    "bg_info",
    "table_bordered",
    "table_banded",
    "table_compact",
    "number_comma",
    "number_precision",
    "percent",
    "currency_usd",
    "currency_eur",
    "date_short",
    "datetime_short",
    "time_short",
]



from __future__ import annotations

import tempfile
from datetime import date, datetime
from pathlib import Path

import openpyxl

import xpyxl as x


def test_xlsxwriter_applies_custom_date_formats() -> None:
    with tempfile.TemporaryDirectory() as tmpdir:
        output_path = Path(tmpdir) / "output.xlsx"
        custom_date = x.Style(name="custom_date", number_format="dd/mm/yyyy")
        custom_datetime = x.Style(
            name="custom_datetime", number_format="dd/mm/yyyy hh:mm"
        )

        workbook = x.workbook()[
            x.sheet("Dates")[
                x.row()[
                    x.cell(style=[custom_date])[date(2025, 1, 2)],
                    x.cell(style=[custom_datetime])[datetime(2025, 1, 2, 14, 30)],
                ]
            ]
        ]

        workbook.save(output_path, engine="xlsxwriter")

        result_wb = openpyxl.load_workbook(output_path)
        ws = result_wb["Dates"]

        assert ws["A1"].number_format == "dd/mm/yyyy"
        assert ws["B1"].number_format == "dd/mm/yyyy hh:mm"
        assert ws["A1"].is_date
        assert ws["B1"].is_date
        assert not isinstance(ws["A1"].value, str)
        assert not isinstance(ws["B1"].value, str)
